import unittest

import numpy as np

from Orange.misc import DistMatrix
from Orange.data import Table, Domain, ContinuousVariable
from orangecontrib.network.widgets.utils import items_from_distmatrix, weights_from_distances


class TestItemsFromMatrix(unittest.TestCase):
    def test_items_from_matrix(self):
        distances = DistMatrix(np.array([[0., 1, 2, 5, 10],
                                         [1, -1, 5, 5, 13],
                                         [2, 5, 2, 6, 13],
                                         [5, 5, 6, 3, 15],
                                         [10, 13, 13, 15, 0]]))

        self.assertEqual(list(items_from_distmatrix(distances).metas[:, 0]),
                         ["1", "2", "3", "4", "5"])

        distances.row_items = list("abcde")
        self.assertEqual(list(items_from_distmatrix(distances).metas[:, 0]),
                         list("abcde"))

        distances.row_items = Table.from_numpy(
            Domain([ContinuousVariable(x) for x in "abcde"]),
            np.arange(25).reshape(5, 5))

        distances.axis = 1
        self.assertIs(items_from_distmatrix(distances), distances.row_items)

        distances.axis = 0
        self.assertEqual(list(items_from_distmatrix(distances).metas[:, 0]),
                         list("abcde"))

    def test_weights_from_distances(self):
        weights = np.array([])
        self.assertEqual(weights_from_distances(weights).size, 0)

        weights = np.array([1])
        np.testing.assert_almost_equal(weights_from_distances(weights), [1])

        weights = np.array([1, 1, 1, 1, 1])
        np.testing.assert_almost_equal( weights_from_distances(weights), [1, 1, 1, 1, 1])

        weights = np.array([0, 0, 0, 0, 0])
        np.testing.assert_almost_equal(weights_from_distances(weights), [1, 1, 1, 1, 1])

        weights = np.array([0, 1])
        np.testing.assert_almost_equal(weights_from_distances(weights), [1, 0.1])

        weights = np.array([2, 3, 4])
        np.testing.assert_almost_equal(weights_from_distances(weights), [1 / 10 ** 0.5, 1 / 10 ** 0.75, 0.1])

        weights = np.array([4, 3, 2])
        np.testing.assert_almost_equal(weights_from_distances(weights), [0.1, 1 / 10 ** 0.75, 1 / 10 ** 0.5])

        weights = np.array([400, 300, 200])
        np.testing.assert_almost_equal(weights_from_distances(weights), [0.1, 1 / 10 ** 0.75, 1 / 10 ** 0.5])

        weights = np.array([400, 300, 200, 0])
        np.testing.assert_almost_equal(weights_from_distances(weights), [0.1, 1 / 10 ** 0.75, 1 / 10 ** 0.5, 1])

        weights = np.array([0.00000007, 0.00000008, 0.00000009])
        np.testing.assert_almost_equal(weights_from_distances(weights), [1, 1, 1])


if __name__ == "__main__":
    unittest.main()



import os
import os.path
import glob

import orange
import orngNetwork

atts = []
atts.append(orange.StringVariable("Network Name"))
atts.append(orange.StringVariable("Network File"))
atts.append(orange.StringVariable("dir"))
atts.append(orange.StringVariable("Item Set"))
atts.append(orange.StringVariable("Edge Set"))
atts.append(orange.FloatVariable("Vertices"))
atts[-1].numberOfDecimals = 0
atts.append(orange.FloatVariable("Edges"))
atts[-1].numberOfDecimals = 0
atts.append(orange.StringVariable("Date"))
atts.append(orange.StringVariable("Description"))

netlist = orange.ExampleTable(orange.Domain(atts, False))

for netFile in glob.glob(os.path.join(os.getcwd(), '*.net')):
	net = orngNetwork.Network.read(netFile)
	name, ext = os.path.splitext(netFile)

	itemFile = ""
	if os.path.exists(name + '_items.tab'):
		itemFile = name + '_items.tab'
	elif os.path.exists(name + '.tab'):
		itemFile = name + '.tab'

	edgeFile = ""
	if os.path.exists(name + '_edges.tab'):
		edgeFile = name + '_edges.tab'

	netlist.append([net.name, os.path.basename(netFile), "doc/datasets/", os.path.basename(itemFile), os.path.basename(edgeFile), net.nVertices, len(net.getEdges()), "4/12/2010", net.description])

netlist.save("network_info.tab")


from typing import Optional

import numpy as np
import scipy.sparse as sp
import pyqtgraph as pg

from AnyQt.QtCore import QLineF, QSize, Qt, Signal, QEvent
from AnyQt.QtGui import QDoubleValidator, QIntValidator, QPalette
from AnyQt.QtWidgets import QApplication, QLineEdit

from Orange.data import Table
from Orange.distance import Euclidean
from Orange.misc import DistMatrix
from Orange.widgets import gui, widget, settings
from Orange.widgets.visualize.utils.plotutils import AxisItem
from Orange.widgets.widget import Input, Output, Msg
from Orange.widgets.utils.widgetpreview import WidgetPreview

from orangecontrib.network.network import Network
from orangecontrib.network.network.base import UndirectedEdges, DirectedEdges
# This enables summarize in widget preview, pylint: disable=unused-import
import orangecontrib.network.widgets
from orangecontrib.network.widgets.utils import items_from_distmatrix, weights_from_distances


class QIntValidatorWithFixup(QIntValidator):
    def fixup(self, text):
        if text and int(text) > self.top():
            return str(self.top())
        return super().fixup(text)


class OWNxFromDistances(widget.OWWidget):
    name = "Network From Distances"
    description = ('Constructs a network by connecting nodes with distances '
                   'below some threshold.')
    icon = "icons/NetworkFromDistances.svg"
    priority = 6440

    class Inputs:
        distances = Input("Distances", DistMatrix)

    class Outputs:
        network = Output("Network", Network)

    resizing_enabled = True
    want_control_area = False

    # The widget stores `density` as setting, because it is more transferable
    # than `threshold`. Internally, the widget uses `threshold` because it is
    # more accurate.
    density = settings.Setting(20, schemaOnly=True)

    class Warning(widget.OWWidget.Warning):
        large_number_of_nodes = \
            Msg('Large number of nodes/edges; expect slow performance.')

    class Error(widget.OWWidget.Error):
        number_of_edges = Msg('The network is too large ({})')

    def __init__(self):
        super().__init__()
        # Matrix from the input, unmodified
        self.matrix: Optional[np.ndarray] = None
        # True, if the matrix is symmetric
        self.symmetric = False
        # All relevant thresholds, that is, all unique values in the distance
        # matrix outside of diagonal. Zero is not explicitly prepended, but
        # will be included if it appears in the distance matrix.
        self.thresholds: Optional[np.ndarray] = None
        # Cumulative frequencies of the thresholds. No prepended zero.
        self.cumfreqs: Optional[np.ndarray] = None
        # Current threshold. Whatever the user sets (threshold, edges, density)
        # will be converted to threshold and stored here.
        self.threshold = 0
        # Number of nodes and edges; use in reports and to set line edits
        self.graph_stat: Optional[tuple[float, float]] = None

        box = gui.vBox(self.mainArea, box=True)
        self.histogram = Histogram(self)
        self.histogram.thresholdChanged.connect(self._on_threshold_dragged)
        self.histogram.draggingFinished.connect(self._on_threshold_drag_finished)
        box.layout().addWidget(self.histogram)

        hbox = gui.hBox(box)
        gui.rubber(hbox)

        _edit_args = dict(alignment=Qt.AlignRight, maximumWidth=50)
        gui.widgetLabel(hbox, "Threshold:")
        self.threshold_edit = QLineEdit(**_edit_args)
        self.threshold_edit.setValidator(QDoubleValidator())
        self.threshold_edit.editingFinished.connect(self._on_threshold_edit)
        hbox.layout().addWidget(self.threshold_edit)
        gui.rubber(hbox)

        gui.widgetLabel(hbox, "Number of edges:")
        self.edges_edit = QLineEdit(**_edit_args)
        self.edges_edit.setValidator(QIntValidatorWithFixup())
        self.edges_edit.editingFinished.connect(self._on_edges_edit)
        hbox.layout().addWidget(self.edges_edit)
        gui.rubber(hbox)

        gui.widgetLabel(hbox, "Density (%):")
        self.density_edit = QLineEdit(**_edit_args)
        self.density_edit.setValidator(QIntValidatorWithFixup(0, 100))
        self.density_edit.editingFinished.connect(self._on_density_edit)
        hbox.layout().addWidget(self.density_edit)
        gui.rubber(hbox)

        self._set_controls_enabled(False)

    def sizeHint(self):  # pragma: no cover
        return QSize(600, 500)

    @property
    def eff_distances(self):
        n = len(self.matrix)
        return n * (n - 1) // (1 + self.symmetric)

    def _on_threshold_dragged(self, threshold):
        self.threshold = threshold
        self.update_edits(from_dragging=True)

    def _on_threshold_drag_finished(self):
        self.generate_network()

    def _on_threshold_edit(self):
        self.threshold = float(self.threshold_edit.text())
        self.generate_network()
        self.update_histogram_lines()
        self.threshold_edit.selectAll()

    def _on_density_edit(self):
        self.density = int(self.density_edit.text())
        self.set_threshold_from_density()
        self.generate_network()
        self.update_histogram_lines()
        self.density_edit.selectAll()

    def _on_edges_edit(self):
        edges = int(self.edges_edit.text())
        self.set_threshold_from_edges(edges)
        self.generate_network()
        self.update_histogram_lines()
        self.edges_edit.selectAll()

    def set_threshold_from_density(self):
        self.set_threshold_from_edges(
            int(np.ceil(self.density * self.eff_distances / 100)))

    def set_threshold_from_edges(self, edges):
        # Set the threshold that will give at least the given number of edges
        if edges == 0:
            self.threshold = 0
        else:
            matrix = self.matrix
            if not self.symmetric:
                matrix = self.__no_diagonal(matrix)
            thresholds = np.sort(matrix.flat)
            edges = min(edges, len(thresholds)) or 1
            self.threshold = thresholds[edges - 1]

    def edges_from_threshold(self):
        """
        Fast, histogram-based estimate of the number of edges below
        the current threshold.
        """
        idx = np.searchsorted(self.thresholds, self.threshold, side='right')
        return self.cumfreqs[idx - 1] if idx else 0

    def update_edits(self, from_dragging=False):
        n_decimals = max(0, -int(np.floor(np.log10(np.max(self.thresholds)))) + 2)
        if from_dragging or self.graph_stat is None:
            edges = self.edges_from_threshold()
        else:
            _, edges = self.graph_stat
        self.density = int(round(100 * edges / self.eff_distances))

        appx = "~" if from_dragging else ""
        self.threshold_edit.setText(f"{self.threshold:.{n_decimals}f}")
        self.edges_edit.setText(appx + str(edges))
        self.density_edit.setText(appx + str(self.density))

    def update_histogram_lines(self):
        if self.graph_stat is None:
            return
        _, edges = self.graph_stat
        self.histogram.update_region(self.threshold, True, True, edges=edges)

    def _set_controls_enabled(self, enabled):
        for edit in (self.threshold_edit, self.edges_edit, self.density_edit):
            edit.setEnabled(enabled)

    # This can be removed when DistMatrix.flat is fixed to include this code
    @staticmethod
    def __no_diagonal(matrix):
        return np.lib.stride_tricks.as_strided(
            matrix.reshape(matrix.size, -1)[1:],
            shape=(matrix.shape[0] - 1, matrix.shape[1]),
            strides=(matrix.strides[0] + matrix.strides[1],
                     matrix.strides[1]),
            writeable=False
        )

    @Inputs.distances
    def set_matrix(self, matrix: DistMatrix):
        if matrix is not None and matrix.size <= 1:
            matrix = None

        self.matrix = matrix
        if matrix is None:
            self.thresholds = None
            self.symmetric = True
            self._set_controls_enabled(False)
            self.histogram.clear_graph()
            self.generate_network()
            return

        self._set_controls_enabled(True)
        self.symmetric = matrix.is_symmetric()
        if not self.symmetric:
            matrix = self.__no_diagonal(matrix)

        if self.eff_distances < 1000:
            self.thresholds, freq = np.unique(matrix.flat, return_counts=True)
        else:
            freq, edges = np.histogram(matrix.flat, bins=1000)
            self.thresholds = edges[1:]
        self.cumfreqs = np.cumsum(freq)
        self.histogram.set_graph(self.thresholds, self.cumfreqs)

        self.edges_edit.validator().setRange(0, self.eff_distances)
        self.set_threshold_from_density()
        self.generate_network()
        self.update_histogram_lines()

    def generate_network(self):
        self.Error.clear()
        self.Warning.clear()
        matrix = self.matrix

        if matrix is None:
            self.graph_stat = None
            self.Outputs.network.send(None)
            return

        nedges = self.edges_from_threshold()
        if nedges > 200000:
            self.Error.number_of_edges(nedges)
            self.graph_stat = None
            self.Outputs.network.send(None)
            return

        mask = np.array(matrix <= self.threshold)
        if self.symmetric:
            mask &= np.tri(*matrix.shape, k=-1, dtype=bool)
        else:
            mask &= ~np.eye(*matrix.shape, dtype=bool)
        if np.sum(mask):
            # Set the weights so that the edge with the smallest distances
            # have the weight of 1, and the edges with the largest distances
            # have the weight of 0.01; the rest are scaled logarithmically
            weights = matrix[mask].astype(float)
            weights = weights_from_distances(weights)
            edges = sp.csr_matrix((weights, mask.nonzero()), shape=matrix.shape)
        else:
            edges = sp.csr_matrix(matrix.shape)
        edge_type = UndirectedEdges if self.symmetric else DirectedEdges
        graph = Network(items_from_distmatrix(self.matrix), edge_type(edges))
        nodes, edges = graph.number_of_nodes(), graph.number_of_edges()
        self.graph_stat = nodes, edges
        self.Warning.large_number_of_nodes(shown=nodes > 3000 or edges > 10000)
        self.Outputs.network.send(graph)
        self.update_edits()

    def send_report(self):
        # We take the threshold from the edit box to have the same number of
        # decimals (the user may have entered a value with more decimals than
        # we'd set within update_edits
        if self.graph_stat is None:
            return

        self.report_items("Settings", [
            ("Threshold", self.threshold_edit.text()),
            ("Density", self.density_edit.text()),
            ("Edges", self.edges_edit.text()),
        ])

        self.report_name("Histogram")
        self.report_plot(self.histogram)

        nodes, edges = self.graph_stat
        self.report_items(
            "Output network",
            [("Vertices", len(self.matrix)),
             ("Edges", f"{edges} "
                       f"({edges / nodes:.2f} per vertex), "
                       f"density: {edges / self.eff_distances:.2%}")])


pg_InfiniteLine = pg.InfiniteLine


class InfiniteLine(pg_InfiniteLine):
    def paint(self, p, *args):
        # From orange3-bioinformatics:OWFeatureSelection.py, thanks to @ales-erjavec
        brect = self.boundingRect()
        c = brect.center()
        line = QLineF(brect.left(), c.y(), brect.right(), c.y())
        t = p.transform()
        line = t.map(line)
        p.save()
        p.resetTransform()
        p.setPen(self.currentPen)
        p.drawLine(line)
        p.restore()

    def hoverEvent(self, ev):  # pragma: no cover
        if ev.isEnter():
            QApplication.setOverrideCursor(
                Qt.SizeVerCursor if self.angle == 0 else Qt.SizeHorCursor)
        elif ev.isExit():
            QApplication.restoreOverrideCursor()
        return super().hoverEvent(ev)

# Patched so that the Histogram's LinearRegionItem works on MacOS
pg.InfiniteLine = InfiniteLine
pg.graphicsItems.LinearRegionItem.InfiniteLine = InfiniteLine


class Histogram(pg.PlotWidget):
    thresholdChanged = Signal(float)
    draggingFinished = Signal()

    def __init__(self, parent, **kwargs):
        axisItems = kwargs.pop("axisItems", None)
        if axisItems is None:
            axisItems = {"left": AxisItem("left"), "bottom": AxisItem("bottom")}
        super().__init__(
            parent, setAspectLocked=True, background=None, axisItems=axisItems,
            **kwargs)
        self.setBackgroundRole(QPalette.Base)
        self.setPalette(QPalette())
        self.__updateScenePalette()

        self.curve = self.plot([0, 1], [0, 0], pen=pg.mkPen('b', width=3))
        self.fill_curve = self.plotItem.plot([0, 1], [0, 0],
            fillLevel=0, pen=pg.mkPen('b', width=4), brush='#02f3')
        self.plotItem.setContentsMargins(12, 12, 12, 12)
        self.plotItem.vb.setMouseEnabled(x=False, y=False)
        # Add another left axis
        self.prop_axis = prop_axis = AxisItem(orientation='right')
        self.plotItem.layout.addItem(prop_axis, 2, 3)
        prop_axis.linkToView(self.plotItem.vb)
        # Set axes titles
        self.plotItem.setLabel('bottom', "Threshold Distance")
        self.plotItem.setLabel('left', "Number of Edges")
        prop_axis.setLabel("Edge Density", units="%")

        line_args = dict(
            movable=True,
            pen=pg.mkPen('k', width=1, style=Qt.DashLine),
            hoverPen=pg.mkPen('k', width=2, style=Qt.DashLine)
        )
        self.hline = pg.InfiniteLine(angle=0, **line_args)
        self.plotItem.addItem(self.hline, ignoreBounds=True)
        self.hline.sigDragged.connect(self._hline_dragged)
        self.hline.sigPositionChangeFinished.connect(self.draggingFinished)

        self.vline = pg.InfiniteLine(angle=90, **line_args)
        self.plotItem.addItem(self.vline, ignoreBounds=True)
        self.vline.sigDragged.connect(self._vline_dragged)
        self.vline.sigPositionChangeFinished.connect(self.draggingFinished)

        self.clear_graph()

    def setScene(self, scene):  # pragma: no cover
        super().setScene(scene)
        self.__updateScenePalette()

    def __updateScenePalette(self):  # pragma: no cover
        scene = self.scene()
        if scene is not None:
            scene.setPalette(self.palette())

    def changeEvent(self, event):  # pragma: no cover
        if event.type() == QEvent.PaletteChange:
            self.__updateScenePalette()
            self.resetCachedContent()
        super().changeEvent(event)

    def update_region(self, thresh,
                      set_hline=False, set_vline=False,
                      edges=None):
        xData, yData = self.curve.xData, self.curve.yData
        high = np.searchsorted(xData, thresh, side='right')
        self.fill_curve.setData(xData[:high], yData[:high])
        if set_hline:
            if edges is None:
                if high == len(yData):
                    edges = yData[-1]
                else:
                    edges = yData[high - (xData[high] > thresh)]
            self.hline.setPos(edges)
        if set_vline:
            self.vline.setPos(thresh)

    def _hline_dragged(self):
        xData, yData = self.curve.xData, self.curve.yData
        pos = self.hline.value()
        idx = np.searchsorted(yData, pos, side='left')
        thresh = xData[min(idx, len(xData) - 1)]
        self.update_region(thresh, set_vline=True)
        self.thresholdChanged.emit(thresh)

    def _vline_dragged(self):
        thresh = self.vline.value()
        self.update_region(thresh, set_hline=True)
        self.thresholdChanged.emit(thresh)

    def _elements(self):
        return self.curve, self.fill_curve, self.hline, self.vline

    def clear_graph(self):
        for el in self._elements():
            el.hide()
        self.prop_axis.setScale(1)

    def set_graph(self, edges, cumfreqs):
        self.curve.setData(np.hstack(([0], edges)),
                           np.hstack(([0], cumfreqs)))
        self.getAxis('left').setRange(0, cumfreqs[-1])
        self.hline.setBounds([0, cumfreqs[-1]])
        self.prop_axis.setScale(1 / cumfreqs[-1] * 100)
        self.getAxis('bottom').setRange(0, edges[-1])
        self.vline.setBounds([0, edges[-1]])
        self.update_region(edges[0], set_hline=True, set_vline=True)
        for el in self._elements():
            el.show()


if __name__ == "__main__":
    distances5 = DistMatrix(np.array([[0., 1, 2, 5, 10],
                                      [1, -1, 5, 5, 13],
                                      [2, 5, 2, 6, 13],
                                      [5, 5, 6, 3, 15],
                                      [10, 13, 13, 15, 0]]))
    distances = Euclidean(Table("iris"))
    WidgetPreview(OWNxFromDistances).run(set_matrix=distances)



import os
import numpy as np
import scipy.sparse as sp

import orangecontrib
from Orange.data import Table

from orangecontrib.network import Network
from orangecontrib.network.network.base import DirectedEdges, UndirectedEdges
from orangecontrib.network.widgets.OWNxFile import OWNxFile
from orangewidget.tests.base import WidgetTest


def _create_net(edges, n=None, directed=False):
    edge_cons = DirectedEdges if directed else UndirectedEdges
    row, col, data = zip(*edges)
    if n is None:
        n = max(*row, *col) + 1
    return Network(np.arange(n), edge_cons(sp.coo_matrix((data, (row, col)), shape=(n, n))))


class NetworkTest(WidgetTest):

    def _read_network(self, filename=None):
        owfile = self.create_widget(OWNxFile)
        owfile.open_net_file(self._get_filename(filename, "n"))
        return self.get_output(owfile.Outputs.network, widget=owfile)

    def _read_items(self, filename=None):
        return Table(self._get_filename(filename))

    def _get_filename(self, filename, mode="d"):
        path = os.path.split(orangecontrib.network.__file__)[0]
        if filename is None:
            path = os.path.join(path, "widgets", "tests")
            filename = "test_items.tab" if mode == "d" else "test.net"
        return os.path.join(path, os.path.join("networks", filename))



from itertools import chain

import numpy as np
from AnyQt.QtWidgets import QFormLayout

from Orange.data import DiscreteVariable, Table
from Orange.widgets import gui
from Orange.widgets.settings import DomainContextHandler, ContextSetting, \
    Setting
from Orange.widgets.utils.itemmodels import VariableListModel
from Orange.widgets.utils.signals import Output, Input
from Orange.widgets.widget import OWWidget, Msg

from orangecontrib.network import Network
from orangecontrib.network.network import twomode


class OWNxSingleMode(OWWidget):
    name = "Single Mode"
    description = "Convert multimodal graphs to single modal"
    icon = "icons/SingleMode.svg"
    priority = 7000

    want_main_area = False
    resizing_enabled = False

    settingsHandler = DomainContextHandler(
        match_values=DomainContextHandler.MATCH_VALUES_ALL)
    variable = ContextSetting(None)
    connect_value = ContextSetting(0)
    connector_value = ContextSetting(0)
    weighting = Setting(0)

    class Inputs:
        network = Input("Network", Network)

    class Outputs:
        network = Output("Network", Network)

    class Warning(OWWidget.Warning):
        ignoring_missing = Msg("Nodes with missing data are being ignored.")

    class Error(OWWidget.Error):
        no_data = Msg("Network has additional data.")
        no_categorical = Msg("Data has no categorical features.")
        same_values = Msg("Values for modes cannot be the same.")

    def __init__(self):
        super().__init__()
        self.network = None

        form = QFormLayout()
        form.setFieldGrowthPolicy(form.AllNonFixedFieldsGrow)
        gui.widgetBox(self.controlArea, box="Mode indicator", orientation=form)
        form.addRow("Feature:", gui.comboBox(
            None, self, "variable", model=VariableListModel(),
            callback=self.indicator_changed))
        form.addRow("Connect:", gui.comboBox(
            None, self, "connect_value",
            callback=self.connect_combo_changed))
        form.addRow("by:", gui.comboBox(
            None, self, "connector_value",
            callback=self.connector_combo_changed))

        gui.comboBox(
            self.controlArea, self, "weighting", box="Edge weights",
            items=[x.name for x in twomode.Weighting],
            callback=self.update_output)

        self.lbout = gui.widgetLabel(gui.hBox(self.controlArea, "Output"), "")
        self._update_combos()
        self._set_output_msg()

    @Inputs.network
    def set_network(self, network):
        self.closeContext()

        self.Warning.clear()
        self.Error.clear()
        if network is not None:
            if not isinstance(network.nodes, Table):
                network = None
                self.Error.no_data()

        self.network = network
        self._update_combos()
        if self.network is not None:
            self.openContext(network.nodes.domain)
        self.update_output()

    def indicator_changed(self):
        """Called on change of indicator variable"""
        self._update_value_combos()
        self.update_output()

    def connect_combo_changed(self):
        cb_connector = self.controls.connector_value
        if not cb_connector.isEnabled():
            self.connector_value = 2 - self.connect_value
        self.update_output()

    def connector_combo_changed(self):
        self.update_output()

    def _update_combos(self):
        """
        Update all three combos

        Set the combo for indicator variable and call the method to update
        combos for values"""
        model = self.controls.variable.model()
        if self.network is None:
            model.clear()
            self.variable = None
        else:
            domain = self.network.nodes.domain
            model[:] = [
                var for var in chain(domain.variables, domain.metas)
                if isinstance(var, DiscreteVariable) and len(var.values) >= 2]
            if not model.rowCount():
                self.Error.no_categorical()
                self.network = None
                self.variable = None
            else:
                self.variable = model[0]
        self._update_value_combos()

    def _update_value_combos(self):
        """Update combos for values"""
        cb_connect = self.controls.connect_value
        cb_connector = self.controls.connector_value
        variable = self.variable

        cb_connect.clear()
        cb_connector.clear()
        cb_connector.setDisabled(variable is None or len(variable.values) == 2)
        self.connect_value = 0
        self.connector_value = 0
        if variable is not None:
            cb_connect.addItems(variable.values)
            cb_connector.addItems(["(all others)"] + list(variable.values))
            self.connector_value = len(variable.values) == 2 and 2

    def update_output(self):
        """Output the network on the output"""
        self.Warning.ignoring_missing.clear()
        self.Error.same_values.clear()
        new_net = None
        if self.network is not None:
            if self.connect_value == self.connector_value - 1:
                self.Error.same_values()
            else:
                mode_mask, conn_mask = self._mode_masks()
                new_net = twomode.to_single_mode(
                    self.network, mode_mask, conn_mask, self.weighting)

        self.Outputs.network.send(new_net)
        self._set_output_msg(new_net)

    def _mode_masks(self):
        """Return indices of nodes in the two modes"""
        data = self.network.nodes
        col_view = data.get_column(self.variable)
        column = col_view.astype(int)
        # Note: conversion required to handle empty (object) arrays
        missing_mask = np.isnan(col_view.astype(float))
        if np.any(missing_mask):
            column[missing_mask] = -1
            self.Warning.ignoring_missing()

        mode_mask = column == self.connect_value
        if self.connector_value:
            conn_mask = column == self.connector_value - 1
        else:
            conn_mask = np.logical_and(column != self.connect_value, np.logical_not(missing_mask))
        return mode_mask, conn_mask

    def _set_output_msg(self, out_network=None):
        if out_network is None:
            self.lbout.setText("No network on output")
        else:
            self.lbout.setText(
                f"{out_network.number_of_nodes()} nodes, "
                f"{out_network.number_of_edges()} edges")

    def send_report(self):
        if self.network:
            self.report_items("", [
                ('Input network',
                 "{} nodes, {} edges".format(
                     self.network.number_of_nodes(),
                     self.network.number_of_edges())),
                ('Mode',
                 self.variable and bool(self.variable.values) and (
                     "Select {}={}, connected through {}".format(
                         self.variable.name,
                         self.variable.values[self.connect_value],
                         "any node" if not self.connector_value
                         else self.variable.values[self.connector_value - 1]
                     ))),
                ('Weighting',
                 bool(self.weighting)
                 and twomode.Weighting[self.weighting].name),
                ("Output network", self.lbout.text())
            ])


def main():  # pragma: no cover
    import OWNxFile
    from AnyQt.QtWidgets import QApplication
    a = QApplication([])
    ow = OWNxSingleMode()
    ow.show()

    def set_network(data, id=None):
        ow.set_network(data)

    owFile = OWNxFile.OWNxFile()
    owFile.Outputs.network.send = set_network
    owFile.open_net_file("/Users/janez/Downloads/100_petrozavodsk_171_events_no_sqrt.net")
    ow.handleNewSignals()
    a.exec_()
    ow.saveSettings()
    owFile.saveSettings()


if __name__ == "__main__":  # pragma: no cover
    main()



import unittest
from unittest.mock import patch

import numpy as np

from Orange.misc import DistMatrix
from Orange.widgets.tests.base import WidgetTest

from orangecontrib.network.widgets.OWNxNeighbor import OWNxNeighbor


class TestOWNxFromDistances(WidgetTest):
    def setUp(self):
        self.widget: OWNxNeighbor = self.create_widget(OWNxNeighbor)

        # Put non-zero elements in the diagonal to test if the widget ignores them
        self.distances = DistMatrix(np.array([[0., 1, 2, 5, 10],
                                              [1, -1, 5, 1, 13],
                                              [2, 5, 2, 6, 11],
                                              [5, 1, 6, 3, 15],
                                              [10, 13, 11, 15, 0]]))

    def test_get_neighbors(self):
        widget = self.widget
        widget.matrix = self.distances

        widget.k = 2
        nearest2 =  [(0, 1), (0, 2),
                     (1, 0), (1, 3),
                     (2, 0), (2, 1),
                     (3, 1), (3, 0),
                     (4, 0), (4, 2)]

        widget.directed = False
        edges = widget.get_neighbors()
        # when sorted, col will be lower, so zip it that way
        pairs = sorted(zip(edges.col, edges.row))
        np.testing.assert_equal(pairs, sorted({tuple(sorted(x)) for x in nearest2}))

        widget.directed = True
        edges = widget.get_neighbors()
        # zip row first; this is how nearest2 is ordered, too
        pairs = sorted(zip(edges.row, edges.col))
        np.testing.assert_equal(pairs, sorted(nearest2))

        widget.k = 1
        widget.directed = False
        edges = widget.get_neighbors()
        pairs = sorted(zip(edges.col, edges.row))
        np.testing.assert_equal(pairs, sorted({tuple(sorted(x)) for x in nearest2[::2]}))

        widget.directed = True
        edges = widget.get_neighbors()
        pairs = sorted(zip(edges.row, edges.col))
        np.testing.assert_equal(pairs, sorted(nearest2[::2]))

        nearest3 = [(0, 1), (0, 2), (0, 3), (1, 0), (1, 2), (1, 3), (2, 0),
                    (2, 1), (2, 3),
                    (3, 0), (3, 1), (3, 2), (4, 0), (4, 1), (4, 2)]

        widget.k = 3
        widget.directed = True
        edges = widget.get_neighbors()
        pairs = sorted(zip(edges.row, edges.col))
        np.testing.assert_equal(pairs, nearest3)

        widget.directed = False
        edges = widget.get_neighbors()
        pairs = sorted(zip(edges.col, edges.row))
        np.testing.assert_equal(pairs,
                                sorted({tuple(sorted(x)) for x in nearest3}))

        widget.k = 4
        widget.directed = True
        self.assertEqual(widget.get_neighbors().nnz, 20)

        widget.directed = False
        self.assertEqual(widget.get_neighbors().nnz, 10)

        widget.k = 6
        widget.directed = True
        self.assertEqual(widget.get_neighbors().nnz, 20)

        widget.directed = False
        self.assertEqual(widget.get_neighbors().nnz, 10)

    def test_output(self):
        nearest2 =  [(0, 1), (0, 2),
                     (1, 0), (1, 3),
                     (2, 0), (2, 1),
                     (3, 1), (3, 0),
                     (4, 0), (4, 2)]

        widget = self.widget
        widget.auto_apply = True
        widget.k = 2
        widget.directed = False
        self.send_signal(self.distances)
        widget.send_report()

        output = self.get_output()
        edges = output.edges[0].edges.tocoo()
        pairs = sorted(zip(edges.col, edges.row))
        np.testing.assert_equal(pairs, sorted({tuple(sorted(x)) for x in nearest2}))

        widget.controls.directed.click()
        output = self.get_output()
        edges = output.edges[0].edges.tocoo()
        pairs = sorted(zip(edges.row, edges.col))
        np.testing.assert_equal(pairs, sorted(nearest2))

    def test_bad_matrix(self):
        widget = self.widget
        widget.auto_apply = True

        self.send_signal(self.distances)
        self.assertIsNotNone(self.get_output())

        self.send_signal(None)
        self.assertIsNone(self.get_output())
        widget.send_report()

        self.send_signal(DistMatrix(np.zeros((0, 0))))
        self.assertIsNone(self.get_output())
        widget.send_report()

        self.send_signal(self.distances)
        self.assertIsNotNone(self.get_output())

        self.send_signal(DistMatrix(np.zeros((0, 0))))
        self.assertIsNone(self.get_output())

        self.send_signal(None)
        self.assertIsNone(self.get_output())

        self.send_signal(DistMatrix([[1]]))
        self.assertIsNone(self.get_output())
        widget.send_report()

        self.send_signal(self.distances)
        self.assertIsNotNone(self.get_output())

        self.send_signal(DistMatrix([[1]]))
        self.assertIsNone(self.get_output())


if __name__ == "__main__":
    unittest.main()


import time

import numpy as np
import pyqtgraph as pg

from AnyQt.QtCore import QLineF, Qt, QRectF
from AnyQt.QtGui import QPen

from Orange.util import scale
from Orange.widgets.settings import Setting
from Orange.widgets.visualize.owscatterplotgraph import OWScatterPlotBase


class PlotVarWidthCurveItem(pg.PlotCurveItem):
    def __init__(self, directed, *args, **kwargs):
        self.directed = directed
        self.__setWidths(kwargs.pop("widths", None))
        self.setPen(kwargs.pop("pen", pg.mkPen(0.0)))
        self.sizes = kwargs.pop("size", None)
        self.coss = self.sins = None
        super().__init__(*args, **kwargs)

    def setWidths(self, widths):
        self.__setWidths(widths)
        self.update()

    def __setWidths(self, widths):
        if widths is not None:
            widths = np.ceil(widths).astype(int)
        self.widths = widths

    def setPen(self, pen):
        self.pen = pen
        self.pen.setCapStyle(Qt.RoundCap)

    def setData(self, *args, **kwargs):
        self.__setWidths(kwargs.pop("widths", self.widths))
        self.setPen(kwargs.pop("pen", self.pen))
        self.sizes = kwargs.pop("size", self.sizes)
        super().setData(*args, **kwargs)

    def paint(self, p, opt, widget):
        def get_arrows():
            # Compute (n, 4) array of coordinates of arrows' ends
            # Arrows are at 15 degrees; length is 10, clipped to edge length
            x0, y0, x1, y1 = edge_coords.T
            arr_len = np.clip(lengths - sizes1 - w3, 0, 10)
            cos12 = arr_len * np.cos(np.pi / 12)
            sin12 = arr_len * np.sin(np.pi / 12)

            # cos(a ± 15) = cos(a) cos(15) ∓ sin(a) sin(15)
            tx = sins * (fx * sin12)
            x1a = x1 - coss * (fx * cos12)
            x2a = x1a - tx
            x1a += tx

            # sin(a ± 15) = sin(a) cos(15) ± sin(15) cos(a)
            ty = (fy * sin12) * coss
            y1a = y1 + sins * (fy * cos12)
            y2a = y1a - ty
            y1a += ty
            return np.vstack((x1a, y1a, x2a, y2a)).T

        def get_short_edge_coords():
            # Compute the target-side coordinates of edges with arrows
            # Such edges are shorted by 8 pixels + width / 3
            off = 8 + w3
            return edge_coords[:, 2:] + (off * np.vstack((-fxcos, fysin))).T

        if self.xData is None or len(self.xData) == 0:
            return

        # Widths of edges, divided by 3; used for adjusting sizes
        w3 = (self.widths if self.widths is not None else self.pen.width()) / 3

        # Sizes of source and target nodes; they are used for adjusting the
        # edge lengths, so we increase the sizes by edge widths / 3
        sizes0, sizes1 = self.sizes[::2] + w3, self.sizes[1::2] + w3

        # Coordinates of vertices for all end points (in real world)
        x0s, x1s = self.xData[::2], self.xData[1::2]
        y0s, y1s = self.yData[::2], self.yData[1::2]

        # Factors for transforming real-worlds coordinates into pixels
        fx = 1 / p.worldTransform().m11()
        fy = 1 / p.worldTransform().m22()

        # Computations of angles (lengths are also used to clip the arrows)
        # We need sine and cosine of angles, and never the actual angles.
        # Sine and cosine are compute as ratios in triangles rather than with
        # trigonometric functions
        diffx, diffy = (x1s - x0s) / fx, -(y1s - y0s) / fy
        lengths = np.sqrt(diffx ** 2 + diffy ** 2)
        arcs = lengths == 0
        coss, sins = np.nan_to_num(diffx / lengths), np.nan_to_num(diffy / lengths)

        # A slower version of the above, with trigonometry
        #  angles = np.arctan2(-(y1s - y0s) / fy, (x1s - x0s) / fx)
        #  return np.cos(angles), np.sin(angles)

        # Sin and cos are mostly used as mulitplied with fx and fy; precompute
        fxcos, fysin = fx * coss, fy * sins

        # Coordinates of edges' end points: coordinates of vertices, adjusted
        # by sizes. When drawing arraws, the target coordinate is used for
        # the tip of the arrow, not the edge
        edge_coords = np.vstack((x0s + fxcos * sizes0, y0s - fysin * sizes0,
                                 x1s - fxcos * sizes1, y1s + fysin * sizes1)).T

        pen = QPen(self.pen)
        p.setRenderHint(p.Antialiasing, True)
        p.setCompositionMode(p.CompositionMode_SourceOver)
        if self.widths is None:
            p.setPen(pen)
            if self.directed:
                for (x0, y0, x1, y1), (x1w, y1w), (xa1, ya1, xa2, ya2), arc in zip(
                        edge_coords, get_short_edge_coords(), get_arrows(), arcs):
                    if not arc:
                        p.drawLine(QLineF(x0, y0, x1w, y1w))
                        p.drawLine(QLineF(xa1, ya1, x1, y1))
                        p.drawLine(QLineF(xa2, ya2, x1, y1))
            else:
                for ecoords in edge_coords[~arcs]:
                    p.drawLine(QLineF(*ecoords))
        else:
            if self.directed:
                for (x0, y0, x1, y1), (x1w, y1w), (xa1, ya1, xa2, ya2), w, arc in zip(
                        edge_coords, get_short_edge_coords(), get_arrows(),
                        self.widths, arcs):
                    if not arc:
                        pen.setWidth(w)
                        p.setPen(pen)
                        p.drawLine(QLineF(x0, y0, x1w, y1w))
                        p.drawLine(QLineF(xa1, ya1, x1, y1))
                        p.drawLine(QLineF(xa2, ya2, x1, y1))
            else:
                for ecoords, w in zip(edge_coords[~arcs], self.widths[~arcs]):
                    pen.setWidth(w)
                    p.setPen(pen)
                    p.drawLine(QLineF(*ecoords))

        # This part is not so optimized because there can't be that many loops
        if np.any(arcs):
            xs, ys = self.xData[::2][arcs], self.yData[1::2][arcs]
            sizes = self.sizes[::2][arcs]
            sizes += w3 if isinstance(w3, float) else w3[arcs]
            # if radius of loop would be size, then distance betwween
            # vertex and loop centers would be
            # d = np.sqrt(size ** 2 - r ** 2 / 2) + r / np.sqrt(2) + r / 2
            ds = sizes * (1 + np.sqrt(2) / 4)
            rxs = xs - ds * fx
            rys = ys - ds * fy
            rfxs = sizes * fx
            rfys = sizes * fy

            ax0o = 6 * np.cos(np.pi * 5 / 6) * fx
            ax1o = 6 * np.cos(np.pi * 7 / 12) * fx
            ay0o = 6 * np.sin(np.pi * 5 / 6) * fy
            ay1o = 6 * np.sin(np.pi * 7 / 12) * fy

            if self.widths is None:
                widths = np.full(len(rxs), pen.width())
            else:
                widths = self.widths[arcs]
            for rx, ry, rfx, rfy, w in zip(rxs, rys, rfxs, rfys, widths):
                rect = QRectF(rx, ry, rfx, rfy)
                pen.setWidth(w)
                p.setPen(pen)
                p.drawArc(rect, 100 * 16, 250 * 16)
                if self.directed:
                    rx += 1.1 * rfx
                    ry += rfy / 2
                    p.drawLine(QLineF(rx, ry, rx + ax0o, ry - ay0o))
                    p.drawLine(QLineF(rx, ry, rx + ax1o, ry - ay1o))


class GraphView(OWScatterPlotBase):
    show_edge_weights = Setting(False)
    relative_edge_widths = Setting(True)
    edge_width = Setting(2)
    label_selected_edges = Setting(True)

    COLOR_NOT_SUBSET = (255, 255, 255, 255)
    COLOR_SUBSET = (0, 0, 0, 255)
    COLOR_DEFAULT = (255, 255, 255, 0)

    class Simplifications:
        NoLabels, NoEdges, NoEdgeLabels, NoDensity, = 1, 2, 4, 8
        NoSimplifications, All = 0, 255

    def __init__(self, master, parent=None):
        super().__init__(master)
        self._reset_attributes()
        self.simplify = self.Simplifications.NoSimplifications
        self.step_resizing.connect(self.update_edges)
        self.end_resizing.connect(self.update_edges)

    def clear(self):
        super().clear()
        self._reset_attributes()

    def _reset_attributes(self):
        self.pair_indices = None
        self.edge_curve = None
        self.edge_labels = []
        self.scatterplot_marked = None
        self.last_click = (-1, None)

    def update_coordinates(self):
        super().update_coordinates()
        self.update_marks()
        self.update_edges()

    def set_simplifications(self, simplifications):
        S = self.Simplifications
        for flag, remove, update in (
                (S.NoDensity, self._remove_density, self.update_density),
                (S.NoLabels, self._remove_labels, self.update_labels),
                (S.NoEdges, self._remove_edges, self.update_edges),
                (S.NoEdgeLabels,
                 self._remove_edge_labels, self.update_edge_labels)):
            if simplifications & flag != self.simplify & flag:
                if simplifications & flag:
                    self.simplify += flag
                    remove()
                else:
                    self.simplify -= flag
                    update()

    def update_edges(self):
        if not self.scatterplot_item \
                or self.simplify & self.Simplifications.NoEdges:
            return
        x, y = self.scatterplot_item.getData()
        edges = self.master.get_edges()
        srcs, dests, weights = edges.row, edges.col, edges.data
        if self.edge_curve is None:
            self.pair_indices = np.empty((2 * len(srcs),), dtype=int)
            self.pair_indices[::2] = srcs
            self.pair_indices[1::2] = dests

        data = dict(x=x[self.pair_indices], y=y[self.pair_indices],
                    pen=self._edge_curve_pen(), antialias=True,
                    size=self.scatterplot_item.data["size"][self.pair_indices] / 2)
        if self.relative_edge_widths and len(set(weights)) > 1:
            data['widths'] = \
                scale(weights, .7, 8) * np.log2(self.edge_width / 4 + 1)
        else:
            data['widths'] = None

        if self.edge_curve is None:
            self.edge_curve = PlotVarWidthCurveItem(
                self.master.is_directed(), **data)
            self.edge_curve.setZValue(-10)
            self.plot_widget.addItem(self.edge_curve)
        else:
            self.edge_curve.setData(**data)
        self.update_edge_labels()

    def set_edge_pen(self):
        if self.edge_curve:
            self.edge_curve.setPen(self._edge_curve_pen())

    def _edge_curve_pen(self):
        return pg.mkPen(
            0.5 if self.class_density else 0.3,
            width=self.edge_width,
            cosmetic=True)

    def update_edge_labels(self):
        for label in self.edge_labels:
            self.plot_widget.removeItem(label)
        self.edge_labels = []
        if self.scatterplot_item is None \
                or not self.show_edge_weights \
                or self.simplify & self.Simplifications.NoEdgeLabels:
            return
        edges = self.master.get_edges()
        if edges is None:
            return
        srcs, dests, weights = edges.row, edges.col, edges.data
        if self.label_selected_edges:
            selected = self._selected_and_marked()
            num_selected = np.sum(selected)
            if num_selected >= 2:
                selected_edges = selected[srcs] & selected[dests]
            else:
                selected_edges = selected[srcs] | selected[dests]
            srcs = srcs[selected_edges]
            dests = dests[selected_edges]
            weights = weights[selected_edges]
        if np.allclose(weights, np.round(weights)):
            labels = [str(x) for x in weights.astype(int)]
        else:
            labels = ["{:.02}".format(x) for x in weights]
        x, y = self.scatterplot_item.getData()
        xs = (x[srcs.astype(np.int64)] + x[dests.astype(np.int64)]) / 2
        ys = (y[srcs.astype(np.int64)] + y[dests.astype(np.int64)]) / 2
        black = pg.mkColor(0, 0, 0)
        for label, x, y in zip(labels, xs, ys):
            ti = pg.TextItem(label, black)
            ti.setPos(x, y)
            self.plot_widget.addItem(ti)
            self.edge_labels.append(ti)

    def _remove_edges(self):
        if self.edge_curve:
            self.plot_widget.removeItem(self.edge_curve)
            self.edge_curve = None
        self._remove_edge_labels()

    def _remove_edge_labels(self):
        for label in self.edge_labels:
            self.plot_widget.removeItem(label)
        self.edge_labels = []

    def update_density(self):
        if not self.simplify & self.Simplifications.NoDensity:
            super().update_density()
            self.set_edge_pen()

    # pylint: disable=access-member-before-definition
    def _remove_density(self):
        if self.density_img:
            self.plot_widget.removeItem(self.density_img)
            self.density_img = None

    def _selected_and_marked(self):
        if self.selection is None:
            selection = np.zeros(len(self.scatterplot_item.data), dtype=bool)
        else:
            selection = np.array(self.selection, dtype=bool)
        marked = self.master.get_marked_nodes()
        if marked is not None:
            selection[marked] = 1
        return selection

    def update_labels(self):
        if self.simplify & self.Simplifications.NoLabels:
            return
        # This is not nice, but let's not add methods to the parent just
        # to support this specific needs of network explorer
        # pylint: disable=access-member-before-definition
        saved_selection = self.selection
        if self.label_only_selected and self.scatterplot_item:
            marked = self.master.get_marked_nodes()
            if marked is not None and len(marked):
                self.selection = self._selected_and_marked()
        super().update_labels()
        self.selection = saved_selection

    def _remove_labels(self):
        # pylint: disable=access-member-before-definition
        for label in self.labels:
            self.plot_widget.removeItem(label)
        self.labels = []

    def update_marks(self):
        if self.scatterplot_marked is None:
            self.scatterplot_marked = pg.ScatterPlotItem([], [])
            self.scatterplot_marked.setZValue(-5)
            self.plot_widget.addItem(self.scatterplot_marked)

        self.update_edge_labels()
        marked = self.master.get_marked_nodes()
        if marked is None:
            self.scatterplot_marked.setData([], [])
            return
        x, y = self.get_coordinates()
        if x is None:  # sanity check; there can be no marked nodes if x is None
            return
        self.scatterplot_marked.setData(
            x[marked], y[marked], size=25,
            pen=pg.mkPen(None), brush=pg.mkBrush("#aff"))

    def select_by_click(self, _, points):
        # Poor man's double click
        indices = [p.data() for p in points]
        last_time, last_indices = self.last_click
        self.last_click = (time.time(), indices)
        if time.time() - last_time < 0.5 and indices == last_indices:
            indices = self.master.get_reachable(indices)
        self.select_by_indices(indices)

    def unselect_all(self):
        super().unselect_all()
        if self.label_selected_edges:
            self.update_edge_labels()

    def _update_after_selection(self):
        if self.label_selected_edges:
            self.update_edge_labels()
        super()._update_after_selection()



import shlex
from itertools import count, repeat

import numpy as np
import scipy.sparse as sp

from Orange.data.io import FileFormatMeta
from .base import Network, EdgeType

__all__ = ("read_pajek", "write_pajek", "PajekReader")


class NetFileFormat(metaclass=FileFormatMeta):
    pass


class PajekReader(NetFileFormat):
    EXTENSIONS = ('.net', )
    DESCRIPTION = 'Pajek network file format'

    @staticmethod
    def read(filename):
        return read_pajek(filename)

    @staticmethod
    def write(filename, network, labels=None):
        return write_pajek(filename, network, labels)


def read_vertices(lines):
    coordinates = np.zeros((len(lines), 2))
    labels = np.full((len(lines)), None, dtype=object)
    id_idx = {}
    for i, line in enumerate(lines):
        node_id, *parts = shlex.split(line)[:4]
        if not parts:
            label = str(node_id)
        else:
            try:  # The format specification was never set in stone, it seems
                label, x, y, *_ = parts + [None, None]
                x, y = float(x), float(y)
                coordinates[i, :2] = (x, y)
            except (TypeError, ValueError):
                if x is not None:
                    label = line.strip().split(maxsplit=1)[1]
        id_idx[node_id] = i
        labels[i] = label
    if np.sum(np.abs(coordinates)) == 0:
        coordinates = None
    return id_idx, labels, coordinates


def read_edges(id_idx, lines, nvertices):
    lines = [(id_idx[v1], id_idx[v2], abs(float(value)))
             for v1, v2, value, *_ in (line.split()[:3] + [1]
                                       for line in lines)]
    v1s, v2s, values = zip(*lines)
    values = np.array(values)
    if values.size and np.all(values == values[0]):
        values = np.lib.stride_tricks.as_strided(
            values[0], (len(values), ), (0, ))
        values.flags.writeable = False
    return sp.coo_matrix((values, (np.array(v1s), np.array(v2s))),
                         shape=(nvertices, nvertices))


def read_edges_list(id_idx, lines, nvertices):
    lines = {id_idx[source]: [id_idx[t] for t in targets]
             for source, *targets in (line.split() for line in lines)}
    # Todo: We can avoid using potentially long Python list by writing directly
    # into np.arrays
    indptr = np.zeros(nvertices + 1, dtype=int)
    indices = []
    for idx in range(nvertices):
        targets = lines.get(idx, ())
        indices += targets
        indptr[idx + 1] = len(indices)
    indices = np.array(indices, dtype=int)
    data = np.lib.stride_tricks.as_strided(np.ones(1), (len(indices), ), (0, ))
    data.flags.writeable = False
    return sp.csr_matrix((data, indices, indptr), shape=(nvertices, nvertices))


def read_pajek(path):
    def check_has_vertices():
        if labels is None:
            raise ValueError("Vertices must be defined before edges or arcs")

    with open(path) as f:
        lines = [line.strip() for line in f]
    lines = np.array([line for line in lines if line and line[0] != "%"])
    part_starts = [(line_no, line)
                   for line_no, line in enumerate(lines)
                   if line[0] == "*"]
    network_name = None
    edges = []
    labels = coordinates = None
    in_first_mode = None
    for (start_line, part_desc), (end_line, *_) \
            in zip(part_starts, part_starts[1:] + [(len(lines), "")]):
        part_type, part_args, *_ = part_desc.split(maxsplit=1) + [""]
        part_type = part_type.lower()
        line_part = lines[start_line + 1: end_line]
        if not line_part.size:
            continue
        if part_type == "*network":
            if network_name is not None:
                raise ValueError(
                    "Pajek files with multiple networks are not supported")
            network_name = part_args.strip()
        elif part_type == "*vertices":
            if labels is not None:
                raise ValueError(
                    "Pajek files with multiple set of vertices are not "
                    "supported")
            id_idx, labels, coordinates = read_vertices(line_part)
            part_args = part_args.split()
            if len(part_args) > 1:
                in_first_mode = int(part_args[1])
        elif part_type in ("*edges", "*arcs"):
            check_has_vertices()
            edges.append(
                EdgeType[part_type=="*arcs"](
                    read_edges(id_idx, line_part, len(labels)),
                    name=part_args.strip() or part_type[1:]))
        elif part_type in ("*edgeslist", "*arcslist"):
            check_has_vertices()
            edges.append(
                EdgeType[part_type=="*arcslist"](
                    read_edges_list(id_idx, line_part, len(labels)),
                    name=part_args.strip() or part_type[1:]))
    network = Network(labels, edges, network_name or "", coordinates)
    if in_first_mode is not None:
        network.in_first_mode = in_first_mode
    return network


def write_pajek(path, network, labels=None):
    if len(network.edges) > 1:
        raise TypeError(
            "This implementation of Pajek format does not support saving "
            "networks with multiple edge types.")
    f = open(path, "wt") if isinstance(path, str) else path
    f.write(f'*Network "{network.name}"\n')
    _write_vertices(f, network, labels)
    if network.edges:
        _write_edges(f, network)
    if f is not path:
        f.close()


def _write_vertices(f, network, labels=None):
    if labels is None:
        labels = network.nodes

    f.write(f"*Vertices\t{network.number_of_nodes()}\n")
    if network.coordinates is not None:
        coords = network.coordinates
    else:
        coords = repeat(())
    for i, label, coordinates in zip(count(start=1), labels, coords):
        f.write(f'{i:6} "{label}"\t' +
                f"{' '.join(f'{c:.4f}' for c in coordinates)}\n")


def _write_edges(f, network):
    edges = network.edges[0]
    f.write("*Arcs\n" if edges.directed else "*Edges\n")
    mat = edges.edges
    if np.all(mat.data == 1):
        for row, rb, re in zip(count(), mat.indptr, mat.indptr[1:]):
            f.write("".join(f"{row + 1:6} {col + 1:6}\n"
                            for col in mat.indices[rb:re]))
    else:
        for row, rb, re in zip(count(), mat.indptr, mat.indptr[1:]):
            f.write("".join(f"{row + 1:6} {mat.indices[i] + 1:6} "
                            f"{mat.data[i]:.6f}\n" for i in range(rb, re)))


# TODO: doesn't belong here if this module is meant as independent from Orange
def transform_data_to_orange_table(network: Network):
    from Orange.data import (
        Table, Domain, ContinuousVariable, DiscreteVariable, StringVariable)
    n_nodes = network.number_of_nodes()
    attrs = []
    if network.coordinates is not None:
        attrs += [ContinuousVariable('x'), ContinuousVariable('y')]
        coords = network.coordinates
    else:
        coords = np.zeros((n_nodes, 0), dtype=float)
    if getattr(network, "in_first_mode", False):
        attrs.append(DiscreteVariable("mode", values=("0", "1")))
        modes = np.ones((n_nodes, 1), dtype=float)
        modes[:network.in_first_mode] = 0
    else:
        modes = np.zeros((n_nodes, 0), dtype=float)
    domain = Domain(attrs, metas=[StringVariable('label ')])
    network.nodes = Table.from_numpy(
        domain,
        np.hstack((coords, modes)),
        metas=network.nodes.reshape(n_nodes, 1))



#cython: boundscheck=False
#cython: wraparound=False
#cython: initializedcheck=False
#cython: embedsignature=True
#cython: infer_types=False
#cython: language_level=3

"""
Fruchterman, T. M. J., Reingold, E. M.,

Graph Drawing by Force-directed Placement.

http://emr.cs.iit.edu/~reingold/force-directed.pdf
"""

cimport numpy as np
import numpy as np

from libc.math cimport sqrt, fabs


ctypedef np.float64_t   double
ctypedef np.int32_t[:]  arr_i1_t
ctypedef double[:, :]   arr_f2_t
ctypedef double[:]      arr_f1_t
ctypedef np.int8_t[:]   arr_b1_t

cdef:
    double GRAVITY = 20


cdef inline double sqr(double x) nogil:
    return x * x


# disp = d / |d| * F_r(d), where d is the vector between the two points and
# F_r(d) = k^2 / |d|.
# Thus dist = d * k^2 / |d|^2. We work with |d| and avoid square roots.
cpdef void repulse(arr_f2_t pos, double k, arr_f2_t disp) nogil:
    cdef:
        Py_ssize_t n_nodes = pos.shape[0]
        Py_ssize_t n_dim = pos.shape[1]
        double dx, dy, f, d2
        Py_ssize_t i, j

        double k2 = sqr(k)

    for i in range(n_nodes):
        for j in range(i):
            dx = pos[i][0] - pos[j][0]
            dy = pos[i][1] - pos[j][1]
            d2 = sqr(dx) + sqr(dy)
            if d2 == 0:
                continue
            f = k2 / d2
            disp[i, 0] += dx * f
            disp[i, 1] += dy * f
            disp[j, 0] -= dx * f
            disp[j, 1] -= dy * f


cpdef void repulse_sampled(arr_f2_t pos, arr_i1_t sample,
                      double k,
                      arr_f2_t disp) nogil:
    cdef:
        Py_ssize_t n_nodes = pos.shape[0]
        Py_ssize_t n_dim = pos.shape[1]
        double dx, dy, f, d2
        Py_ssize_t i, j, s

        Py_ssize_t sample_size = sample.shape[0]
        double inv_sample_ratio = n_nodes / sample_size
        double k2 = sqr(k)

    for i in range(n_nodes):
        for s in range(sample_size):
            j = sample[s]
            dx = pos[i][0] - pos[j][0]
            dy = pos[i][1] - pos[j][1]
            d2 = sqr(dx) + sqr(dy)
            if d2 == 0:
                continue
            f = k2 / d2 * inv_sample_ratio
            disp[i, 0] += dx * f
            disp[i, 1] += dy * f


cpdef void attract(arr_f2_t pos,
                   arr_i1_t edge_src,  arr_i1_t edge_dst, arr_f1_t edge_weights,
                   double k,
                   arr_f2_t disp) nogil:
    cdef:
        Py_ssize_t n_edges = edge_src.shape[0]

        double dx, dy, f, mag2, weight
        Py_ssize_t i, j, e

        double inv_k = 1 / k
        int weighted = edge_weights.shape[0] != 0

    # TODO are weights in any way normalized?!
    for e in range(n_edges):
        i, j = edge_src[e], edge_dst[e],
        dx = pos[i][0] - pos[j][0]
        dy = pos[i][1] - pos[j][1]
        mag2 = sqr(dx) + sqr(dy)
        f = sqrt(mag2) * inv_k
        if weighted:
            f *= edge_weights[e]
        dx *= f
        dy *= f
        disp[i, 0] -= dx
        disp[i, 1] -= dy
        disp[j, 0] += dx
        disp[j, 1] += dy


cpdef void gravity(arr_f2_t pos, double k, arr_f2_t disp) nogil:
    cdef:
        double dx, dy, f, mag, weight
        Py_ssize_t i

    for i in range(pos.shape[0]):
        """
        Circular gravity; gravity along axes (below) seems to work better
        f = k * GRAVITY * sqrt(sqr(pos[i, 0]) + sqr(pos[i, 1]))
        disp[i, 0] -= f * pos[i, 0]
        disp[i, 1] -= f * pos[i, 1]
        """
        f = k * GRAVITY
        disp[i, 0] -= f * fabs(pos[i, 0]) * pos[i, 0]
        disp[i, 1] -= f * fabs(pos[i, 1]) * pos[i, 1]


cpdef void move(arr_f2_t pos, arr_f2_t disp, double temp) nogil:
    cdef:
        double mag, f
        Py_ssize_t i

        double temp2 = sqr(temp)

    for i in range(pos.shape[0]):
        mag = sqr(disp[i, 0]) + sqr(disp[i, 1])
        if mag == 0:
            continue
        if mag > temp2:
            f = temp / sqrt(mag)
            pos[i, 0] += disp[i, 0] * f
            pos[i, 1] += disp[i, 1] * f
        else:
            pos[i, 0] += disp[i, 0]
            pos[i, 1] += disp[i, 1]


def fruchterman_reingold_step(arr_f2_t pos,
                              arr_i1_t sample,
                              arr_i1_t edge_src,
                              arr_i1_t edge_dst,
                              arr_f1_t edge_weights,
                              double k,
                              double temp,
                              # preallocated for efficiency; same shape as pos
                              arr_f2_t disp):
    cdef:
        double mag, adj, weight
        Py_ssize_t row, col, i, j, d, s
        Py_ssize_t n_nodes = pos.shape[0]
        Py_ssize_t n_dim = pos.shape[1]

    with nogil:
        disp[:, :] = 0
        attract(pos, edge_src, edge_dst, edge_weights, k, disp)
        if  sample.shape[0] == 0:
            repulse(pos, k, disp)
        else:
            repulse_sampled(pos, sample, k, disp)

        gravity(pos, k, disp)
        move(pos, disp, temp)



from AnyQt.QtCore import Qt, QThread
from Orange.data import Table
from Orange.widgets.widget import OWWidget
from gensim.models.callbacks import CallbackAny2Vec
from orangewidget import gui, settings
from orangewidget.utils.signals import Input, Output

from orangecontrib.network import Network
from orangecontrib.network.network import embeddings, readwrite
from orangewidget.utils.widgetpreview import WidgetPreview


class EmbedderThread(QThread):
    def __init__(self, func):
        super().__init__()
        self.func = func
        self.result = None

    def __del__(self):
        self.quit()

    def run(self):
        self.result = self.func()


class ProgressBarUpdater(CallbackAny2Vec):
    def __init__(self, widget, num_epochs):
        self.widget = widget
        self.curr_epoch = 0
        self.num_epochs = num_epochs

    def on_epoch_begin(self, model):
        if self.widget is None:
            return

        self.widget.progressBarSet(100 * (self.curr_epoch / self.num_epochs))

    def on_epoch_end(self, model):
        self.curr_epoch += 1


class OWNxEmbedding(OWWidget):
    name = "Network Embeddings"
    description = "Embed network elements"
    icon = "icons/NetworkEmbedding.svg"
    priority = 6450

    class Inputs:
        network = Input("Network", Network, default=True)

    class Outputs:
        items = Output("Items", Table)

    resizing_enabled = False
    want_main_area = False

    p = settings.Setting(1.0)
    q = settings.Setting(1.0)
    walk_len = settings.Setting(50)
    num_walks = settings.Setting(10)
    emb_size = settings.Setting(128)
    window_size = settings.Setting(5)
    num_epochs = settings.Setting(1)

    auto_commit = settings.Setting(True)

    def __init__(self):
        super().__init__()
        self.network = None
        self.embedder = None
        self._worker_thread = None
        self._progress_updater = None

        def commit():
            return self.commit()

        box = gui.widgetBox(self.controlArea, box=True)
        kwargs = dict(controlWidth=75, alignment=Qt.AlignRight, callback=commit)
        gui.spin(box, self, "p", 0.0, 10.0, 0.1, label="Return parameter (p): ",
                 spinType=float, **kwargs)
        gui.spin(box, self, "q", 0.0, 10.0, 0.1, label="In-out parameter (q): ",
                 spinType=float, **kwargs)
        gui.spin(box, self, "walk_len", 1, 100_000, 1, label="Walk length: ",
                 **kwargs)
        gui.spin(box, self, "num_walks", 1, 10_000, 1, label="Walks per node: ",
                 **kwargs)
        gui.spin(box, self, "emb_size", 1, 10_000, 1, label="Embedding size: ",
                 **kwargs)
        gui.spin(box, self, "window_size", 1, 20, 1, label="Context size: ",
                 **kwargs)
        gui.spin(box, self, "num_epochs", 1, 100, 1, label="Number of epochs: ",
                 **kwargs)

        gui.auto_commit(self.controlArea, self, "auto_commit", "Commit",
                        checkbox_label="Auto-commit", orientation=Qt.Horizontal)
        commit()

    @Inputs.network
    def set_network(self, net):
        self.network = net
        self.commit()

    def commit(self):
        self.Warning.clear()

        # cancel existing computation if running
        if self._worker_thread is not None:
            self._worker_thread.finished.disconnect()
            self._worker_thread.quit()
            self._worker_thread = None

        if self.network is None:
            self.Outputs.items.send(None)
            return

        self._progress_updater = ProgressBarUpdater(self, self.num_epochs)
        self.embedder = embeddings.Node2Vec(self.p, self.q, self.walk_len, self.num_walks,
                                            self.emb_size, self.window_size, self.num_epochs,
                                            callbacks=[self._progress_updater])

        self._worker_thread = EmbedderThread(lambda: self.embedder(self.network))
        self._worker_thread.finished.connect(self.on_finished)
        self.progressBarInit()
        self.progressBarSet(1e-5)
        self._worker_thread.start()

    def on_finished(self):
        output = self._worker_thread.result
        self._worker_thread = None
        self._progress_updater = None
        self.progressBarFinished()
        self.Outputs.items.send(output)

    def onDeleteWidget(self):
        if self._worker_thread is not None:
            # prevent the callback from trying to access deleted widget object
            self._progress_updater.widget = None
            self._worker_thread.finished.disconnect()
            self._worker_thread.quit()
        super().onDeleteWidget()


if __name__ == "__main__":
    from os.path import join, dirname
    davis = join(dirname(__file__), "..", "networks", "davis.net")
    network = readwrite.read_pajek(davis)
    WidgetPreview(OWNxEmbedding).run(network)




import unittest
from unittest.mock import Mock

import numpy as np

import Orange
from orangewidget.tests.utils import simulate

from orangecontrib.network.widgets.tests.utils import NetworkTest
from orangecontrib.network import Network
from orangecontrib.network.widgets.OWNxExplorer import OWNxExplorer


class TestOWNxExplorer(NetworkTest):
    def setUp(self):
        self.widget = self.create_widget(OWNxExplorer)  # type: OWNxExplorer
        self.network = self._read_network("lastfm.net")
        self.data = self._read_items("lastfm.tab")

        self.davis_net = self._read_network("davis.net")
        self.davis_data = self._read_items("davis.tsv")

    def test_minimum_size(self):
        # Disable this test from the base test class
        pass


class TestOWNxExplorerWithLayout(TestOWNxExplorer):
    def test_empty_network(self):
        net = Network([], [])
        # should not crash
        self.send_signal(self.widget.Inputs.network, net)


class TestOWNxEplorerWithoutLayout(TestOWNxExplorer):
    def setUp(self):
        super().setUp()
        self.widget.relayout = Mock()

    def test_too_many_labels(self):
        # check that Warning is shown when attribute
        # with too many labels is chosen
        # GH - 120
        self.send_signal(self.widget.Inputs.network, self.network)
        self.send_signal(self.widget.Inputs.node_data, self.data)
        simulate.combobox_activate_item(self.widget.controls.attr_label,
                                        self.data.domain.metas[0].name)
        self.assertTrue(self.widget.Warning.too_many_labels.is_shown())
        simulate.combobox_activate_index(self.widget.controls.attr_label, 0)
        self.assertFalse(self.widget.Warning.too_many_labels.is_shown())

    def test_subset_selection(self):
        # test if selecting from the graph works

        self.send_signal(self.widget.Inputs.network, self.network)
        self.assertEqual(self.widget.nSelected, 0)
        self.send_signal(self.widget.Inputs.node_data, self.data)
        self.widget.graph.selection_select(np.arange(0, 5))
        outputs = self.widget.Outputs
        self.assertIsInstance(self.get_output(outputs.subgraph), Network)
        self.assertEqual(self.widget.nSelected, 5)

    def test_get_reachable(self):
        # gene label indices, which are equal to their numbers assigned in the .net file - 1
        GENES = {"IDS": 70, "GNS": 36, "BLVRB": 61, "HMOX1": 5, "BLVRA": 62,
                 "PSMA2": 6, "PSMA4": 8, "PSMA5": 9, "PSMA6": 7}
        self.send_signal(self.widget.Inputs.network, self._read_network("leu_by_genesets.net"))

        self.assertSetEqual(set(self.widget.get_reachable([GENES["PSMA2"]])),
                            {GENES["PSMA2"], GENES["PSMA4"], GENES["PSMA5"], GENES["PSMA6"]})
        # test that zero-weight edges do not get dropped due to sparse matrix representation
        self.assertSetEqual(set(self.widget.get_reachable([GENES["IDS"]])),
                            {GENES["IDS"], GENES["GNS"]})
        self.assertSetEqual(set(self.widget.get_reachable([GENES["BLVRB"]])),
                            {GENES["BLVRB"], GENES["HMOX1"], GENES["BLVRA"]})

    def test_edge_weights(self):
        self.send_signal(self.widget.Inputs.network, self.davis_net)
        self.widget.graph.show_edge_weights = True

        # Mark nodes with many connections (multiple): should show the weights for edges between marked nodes only
        self.widget.mark_min_conn = 8
        self.widget.set_mark_mode(9)
        self.assertEqual(len(self.widget.graph.edge_labels), 12)

        # Reset to default (no selection) and check that the labels disappear
        self.widget.set_mark_mode(0)
        self.assertEqual(len(self.widget.graph.edge_labels), 0)

        # Mark nodes with most connections (single): should show all its edges' weights
        self.widget.set_mark_mode(10)
        self.assertEqual(len(self.widget.graph.edge_labels), 14)

    def test_input_subset(self):
        self.send_signal(self.widget.Inputs.network, self.davis_net)
        self.send_signal(self.widget.Inputs.node_data, self.davis_data)
        sub_mask = self.widget.get_subset_mask()
        self.assertIsNone(sub_mask)

        self.send_signal(self.widget.Inputs.node_subset, self.davis_data[:3])
        sub_mask = self.widget.get_subset_mask()
        num_subset_nodes = np.sum(sub_mask)
        self.assertEqual(num_subset_nodes, 3)

        self.send_signal(self.widget.Inputs.node_subset, None)
        sub_mask = self.widget.get_subset_mask()
        self.assertIsNone(sub_mask)

    def test_report(self):
        self.widget.send_report()

        self.send_signal(self.widget.Inputs.network, self.davis_net)
        self.widget.send_report()

        self.send_signal(self.widget.Inputs.node_data, self.davis_data)
        self.widget.send_report()


if __name__ == "__main__":
    unittest.main()



from operator import itemgetter
from os import path
from itertools import product
from traceback import format_exception_only

import numpy as np

from AnyQt.QtCore import Qt
from AnyQt.QtWidgets import QStyle, QSizePolicy, QFileDialog

from Orange.util import get_entry_point
from Orange.data import Table, Domain, StringVariable
from Orange.data.util import get_unique_names
from Orange.widgets import gui, settings
from Orange.widgets.settings import ContextHandler
from Orange.widgets.utils.itemmodels import VariableListModel
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.widget import OWWidget, Msg, Input, Output
from orangecontrib.network.network import Network
from orangecontrib.network.network.readwrite import read_pajek


class NxFileContextHandler(ContextHandler):
    def new_context(self, useful_vars):
        context = super().new_context()
        context.useful_vars = {var.name for var in useful_vars}
        context.label_variable = None
        return context

    # noinspection PyMethodOverriding
    def match(self, context, useful_vars):
        useful_vars = {var.name for var in useful_vars}
        if context.useful_vars == useful_vars:
            return self.PERFECT_MATCH
        # context.label_variable can also be None; this would always match,
        # so ignore it
        elif context.label_variable in useful_vars:
            return self.MATCH
        else:
            return self.NO_MATCH

    def settings_from_widget(self, widget, *_):
        context = widget.current_context
        if context is not None:
            context.label_variable = \
                widget.label_variable and widget.label_variable.name

    def settings_to_widget(self, widget, useful_vars):
        context = widget.current_context
        widget.label_variable = None
        if context.label_variable is not None:
            for var in useful_vars:
                if var.name == context.label_variable:
                    widget.label_variable = var
                    break


demos_path = next(
    get_entry_point("Orange3-Network", "orange.data.io.search_paths", "network")
    ())[1]


class OWNxFile(OWWidget):
    name = "Network File"
    description = "Read network graph file"
    icon = "icons/NetworkFile.svg"
    priority = 6410

    class Inputs:
        items = Input("Items", Table)

    class Outputs:
        network = Output("Network", Network)
        items = Output("Items", Table)

    settingsHandler = NxFileContextHandler()
    label_variable: StringVariable = settings.ContextSetting(None)
    recentFiles = settings.Setting([])

    class Information(OWWidget.Information):
        auto_annotation = Msg(
            'Nodes annotated with data from file with the same name')
        suggest_annotation = Msg(
            'Add optional data input to annotate nodes')

    class Error(OWWidget.Error):
        io_error = Msg('Error reading file "{}"\n{}')
        error_parsing_file = Msg('Error reading file "{}"')
        auto_data_failed = Msg(
            "Attempt to read {} failed\n"
            "The widget tried to annotated nodes with data from\n"
            "a file with the same name.")
        mismatched_lengths = Msg(
            "Data size does not match the number of nodes.\n"
            "Select a data column whose values can be matched with network "
            "labels")

    want_main_area = False
    mainArea_width_height_ratio = None

    def __init__(self):
        super().__init__()

        self.network = None
        self.auto_data = None
        self.original_nodes = None
        self.data = None
        self.net_index = 0

        hb = gui.widgetBox(self.controlArea, orientation=Qt.Horizontal)
        self.filecombo = gui.comboBox(
            hb, self, "net_index", callback=self.select_net_file,
            minimumWidth=250)
        gui.button(
            hb, self, '...', callback=self.browse_net_file, disabled=0,
            icon=self.style().standardIcon(QStyle.SP_DirOpenIcon),
            sizePolicy=(QSizePolicy.Maximum, QSizePolicy.Fixed))
        gui.button(
            hb, self, 'Reload', callback=self.reload,
            icon=self.style().standardIcon(QStyle.SP_BrowserReload),
            sizePolicy=(QSizePolicy.Maximum, QSizePolicy.Fixed))

        self.label_model = VariableListModel(placeholder="(Match by rows)")
        self.label_model[:] = [None]
        gui.comboBox(
            self.controlArea, self, "label_variable", box=True,
            label="Match node labels to data column: ", orientation=Qt.Horizontal,
            model=self.label_model, callback=self.label_changed)

        self.populate_comboboxes()
        self.setFixedHeight(self.sizeHint().height())
        self.reload()

    @Inputs.items
    def set_data(self, data):
        self.data = data
        self.update_label_combo()
        self.send_output()

    def populate_comboboxes(self):
        self.filecombo.clear()
        for file in self.recentFiles or ("(None)",):
            self.filecombo.addItem(path.basename(file))
        self.filecombo.addItem("Browse documentation networks...")
        self.filecombo.updateGeometry()

    def browse_net_file(self, browse_demos=False):
        """user pressed the '...' button to manually select a file to load"""
        if browse_demos:
            startfile = demos_path
        else:
            startfile = self.recentFiles[0] if self.recentFiles else '.'

        filename, _ = QFileDialog.getOpenFileName(
            self, 'Open a Network File', startfile,
            ';;'.join(("Pajek files (*.net *.pajek)",)))
        if not filename:
            return False

        if filename in self.recentFiles:
            self.recentFiles.remove(filename)
        self.recentFiles.insert(0, filename)

        self.populate_comboboxes()
        self.net_index = 0
        self.select_net_file()
        return True

    def reload(self):
        if self.recentFiles:
            self.select_net_file()

    def select_net_file(self):
        """user selected a graph file from the combo box"""
        if self.net_index > len(self.recentFiles) - 1:
            if not self.browse_net_file(True):
                return  # Cancelled
        elif self.net_index:
            self.recentFiles.insert(0, self.recentFiles.pop(self.net_index))
            self.net_index = 0
            self.populate_comboboxes()
        if self.recentFiles:
            self.open_net_file(self.recentFiles[0])

    def open_net_file(self, filename):
        """Read network from file."""
        self.Error.clear()
        self.Warning.clear()
        self.Information.clear()
        self.network = None
        self.original_nodes = None
        try:
            self.network = read_pajek(filename)
        except OSError as err:
            self.Error.io_error(
                filename,
                "".join(format_exception_only(type(err), err)).rstrip())
        except Exception:  # pylint: disable=broad-except
            self.Error.error_parsing_file(filename)
        else:
            self.original_nodes = self.network.nodes
            self.read_auto_data(filename)
        self.update_label_combo()
        self.send_output()

    def read_auto_data(self, filename):
        self.Error.auto_data_failed.clear()

        self.auto_data = None
        errored_file = None
        basenames = (filename,
                     path.splitext(filename)[0],
                     path.splitext(filename)[0] + '_items')
        for basename, ext in product(basenames, ('.tab', '.tsv', '.csv')):
            filename = basename + ext
            if path.exists(filename):
                try:
                    self.auto_data = Table.from_file(filename)
                    break
                except Exception:  # pylint: disable=broad-except
                    errored_file = filename
        else:
            if errored_file:
                self.Error.auto_data_failed(errored_file)

    def update_label_combo(self):
        self.closeContext()
        data = self.data if self.data is not None else self.auto_data
        if self.network is None or data is None:
            self.label_model[:] = [None]
        else:
            best_var, useful_vars = self._vars_for_label(data)
            self.label_model[:] = [None] + useful_vars
            self.label_variable = best_var
            self.openContext(useful_vars)
        self.set_network_nodes()

    def _vars_for_label(self, data: Table):
        vars_and_overs = []
        original_nodes = set(self.original_nodes)
        for var in data.domain.metas:
            if not isinstance(var, StringVariable):
                continue
            values= data.get_column(var)
            values = values[values != ""]
            set_values = set(values)
            # values have to be unique, and have to include all labels
            if len(values) != len(set_values) \
                    or not original_nodes <= set_values:
                continue
            vars_and_overs.append((len(set_values), var))
        if not vars_and_overs:
            return None, []
        # Prefer variables with less extra values
        _, best_var = min(vars_and_overs, key=itemgetter(0))
        useful_string_vars = [var for _, var in vars_and_overs]
        return best_var, useful_string_vars

    def label_changed(self):
        self.set_network_nodes()
        self.send_output()

    def send_output(self):
        if self.network is None:
            self.Outputs.network.send(None)
            self.Outputs.items.send(None)
        else:
            self.Outputs.network.send(self.network)
            self.Outputs.items.send(self.network.nodes)

    def set_network_nodes(self):
        self.Error.mismatched_lengths.clear()
        self.Information.auto_annotation.clear()
        self.Information.suggest_annotation.clear()
        if self.network is None:
            return

        data = self.data if self.data is not None else self.auto_data
        if data is None:
            self.Information.suggest_annotation()
        elif self.label_variable is None \
                and len(data) != self.network.number_of_nodes():
            self.Error.mismatched_lengths()
            data = None

        if data is None:
            self.network.nodes = self._label_to_tabel()
        elif self.label_variable is None:
            self.network.nodes = self._combined_data(data)
        else:
            self.network.nodes = self._data_by_labels(data)

    def _data_by_labels(self, data):
            data_col = data.get_column(self.label_variable)
            data_rows = {label: row for row, label in enumerate(data_col)}
            indices = [data_rows[label] for label in self.original_nodes]
            return data[indices]

    def _combined_data(self, source):
        nodes = np.array(self.original_nodes, dtype=str)
        if nodes.ndim != 1:
            return source
        try:
            nums = np.sort(np.array([int(x) for x in nodes]))
        except ValueError:
            pass
        else:
            if np.all(nums[1:] - nums[:-1] == 1):
                return source

        src_dom = source.domain
        label_attr = StringVariable(get_unique_names(src_dom, "node_label"))
        domain = Domain(src_dom.attributes, src_dom.class_vars,
                        src_dom.metas + (label_attr, ))
        data = source.transform(domain)
        with data.unlocked(data.metas):
            data.metas[:, -1] = nodes
        return data

    def _label_to_tabel(self):
        domain = Domain([], [], [StringVariable("node_label")])
        n = len(self.original_nodes)
        data = Table.from_numpy(
            domain, np.empty((n, 0)), np.empty((n, 0)),
            np.array(self.original_nodes, dtype=str).reshape(-1, 1))
        return data


    def send_report(self):
        if not self.network:
            return

        self.report_items(
            "Network file",
            [("File name", self.filecombo.currentText()),
             ("Vertices", self.network.number_of_nodes()),
             ("Directed", ["No", "Yes"][self.network.edges[0].directed])
             ])


if __name__ == "__main__":
    WidgetPreview(OWNxFile).run()



import unittest
from unittest.mock import Mock, patch

import numpy as np
import scipy.sparse as sp

from Orange.data import Domain, DiscreteVariable, ContinuousVariable, Table

from orangecontrib.network import Network
from orangecontrib.network.widgets.ownxsinglemode import OWNxSingleMode
from orangecontrib.network.widgets.tests.utils import NetworkTest


class TestOWNxSingleMode(NetworkTest):
    def setUp(self):
        self.widget = self.create_widget(OWNxSingleMode)  # type: OWNxSingleMode

        self.a, self.b, self.c = [
            DiscreteVariable("a", values=("a0", "a1")),
            ContinuousVariable("b"),
            DiscreteVariable("c", values=("c0", "c1", "c2", "c3"))]
        self.domain = Domain([self.a, self.b, self.c])
        self.table = Table.from_numpy(self.domain, np.array([
            [0, 0, 2],
            [1, 0, 3],
            [1, 0, 1],
            [0, 2, 1],
            [0, 0, 1]
        ]))

        self.d = DiscreteVariable("d", values=("d0", ))

    def _set_graph(self, data, edges=None):
        net = Network(
            data,
            sp.csr_matrix((len(data), len(data))) if edges is None else edges)
        self.send_signal(self.widget.Inputs.network, net)

    def test_combo_inits(self):
        widget = self.widget
        model = widget.controls.variable.model()
        cb_select = widget.controls.connect_value
        cb_connect = widget.controls.connector_value

        self.assertSequenceEqual(model, [])
        self.assertIsNone(widget.variable)
        self.assertEqual(cb_select.count(), 0)
        self.assertEqual(cb_connect.count(), 0)
        self.assertFalse(widget.Error.no_data.is_shown())
        self.assertFalse(widget.Error.no_categorical.is_shown())
        self.assertFalse(widget.Error.same_values.is_shown())

        self._set_graph(self.table)
        model = widget.controls.variable.model()
        self.assertSequenceEqual(model, [self.a, self.c])
        self.assertIs(widget.variable, self.a)
        self.assertEqual(cb_select.count(), 2)
        self.assertEqual(cb_select.itemText(0), "a0")
        self.assertEqual(cb_select.itemText(1), "a1")
        self.assertEqual(cb_connect.count(), 3)
        self.assertEqual(cb_connect.itemText(0), "(all others)")
        self.assertEqual(cb_connect.itemText(1), "a0")
        self.assertEqual(cb_connect.itemText(2), "a1")

        self.send_signal(widget.Inputs.network, None)
        self.assertSequenceEqual(model, [])
        self.assertIsNone(widget.variable)
        self.assertEqual(cb_select.count(), 0)
        self.assertEqual(cb_connect.count(), 0)
        self.assertFalse(widget.Error.no_data.is_shown())
        self.assertFalse(widget.Error.no_categorical.is_shown())
        self.assertFalse(widget.Error.same_values.is_shown())

        self._set_graph(Table.from_domain(Domain([], [], [self.a, self.c])))
        self.assertSequenceEqual(model, [self.a, self.c])
        self.assertIs(widget.variable, self.a)

    def test_no_single_valued_vars(self):
        self._set_graph(Table.from_domain(Domain([self.a, self.b, self.c, self.d])))

    def test_show_errors(self):
        widget = self.widget
        model = widget.controls.variable.model()
        a, b, c, d = self.a, self.b, self.c, self.d
        cb_connector = widget.controls.connector_value

        no_data = widget.Error.no_data.is_shown
        no_categorical = widget.Error.no_categorical.is_shown
        same_values = widget.Error.same_values.is_shown

        self._set_graph(Table.from_domain(Domain([a, b, c, d])))
        self.assertSequenceEqual(model, [a, c])
        self.assertFalse(no_data())
        self.assertFalse(no_categorical())
        self.assertFalse(same_values())

        self._set_graph(Table.from_domain(Domain([b, d])))
        self.assertSequenceEqual(model, [])
        self.assertFalse(no_data())
        self.assertTrue(no_categorical())
        self.assertFalse(same_values())

        self._set_graph(Table.from_domain(Domain([a, b, c, d])))
        self.assertSequenceEqual(model, [a, c])
        self.assertFalse(no_data())
        self.assertFalse(no_categorical())
        self.assertFalse(same_values())

        widget.connector_value = widget.connect_value + 1
        cb_connector.activated[int].emit(widget.connector_value)
        self.assertFalse(no_data())
        self.assertFalse(no_categorical())
        self.assertTrue(same_values())

        net = Network(range(3), sp.csr_matrix([[0, 1, 0], [1, 2, 0], [0, 0, 0]]))
        self.send_signal(widget.Inputs.network, net)
        self.assertTrue(no_data())
        self.assertFalse(no_categorical())
        self.assertFalse(same_values())

        self._set_graph(Table.from_domain(Domain([a, b, c, d])))
        widget.connector_value = widget.connect_value + 1
        self.send_signal(widget.Inputs.network, None)
        self.assertFalse(no_data())
        self.assertFalse(no_categorical())
        self.assertFalse(same_values())

        self._set_graph(Table.from_domain(Domain([a, b, c, d])))
        widget.connector_value = widget.connect_value + 1
        cb_connector.activated[int].emit(widget.connector_value)
        self.assertFalse(no_data())
        self.assertFalse(no_categorical())
        self.assertTrue(same_values())

        self._set_graph(Table.from_domain(Domain([b, d])))
        self.assertFalse(no_data())
        self.assertTrue(no_categorical())
        self.assertFalse(same_values())

    def test_value_combo_updates(self):
        widget = self.widget
        widget.update_output = Mock()
        cb_kept = widget.controls.connect_value
        a, c = self.a, self.c

        self._set_graph(Table.from_domain(Domain([a, c])))
        self.assertEqual(len(cb_kept), 2)
        widget.update_output.assert_called()
        widget.update_output.reset_mock()

        widget.variable = c
        widget.controls.variable.activated[int].emit(1)
        self.assertEqual(len(cb_kept), 4)
        widget.update_output.assert_called()
        widget.update_output.reset_mock()

        widget.connect_value = 3
        widget.variable = a
        widget.controls.variable.activated[int].emit(0)
        self.assertEqual(len(cb_kept), 2)
        self.assertEqual(widget.connect_value, 0)
        widget.update_output.assert_called()
        widget.update_output.reset_mock()

    def test_disable_combo_on_binary(self):
        widget = self.widget
        a, c = self.a, self.c
        cb_variable = widget.controls.variable
        cb_connect = widget.controls.connect_value
        cb_connector = widget.controls.connector_value

        self.assertFalse(cb_connector.isEnabled())

        self._set_graph(Table.from_domain(Domain([a, c])))
        self.assertFalse(cb_connector.isEnabled())
        self.assertEqual(widget.connector_value, 2)

        widget.connect_value = 1
        cb_connect.activated[int].emit(1)
        self.assertFalse(cb_connector.isEnabled())
        self.assertEqual(widget.connector_value, 1)

        widget.variable = c
        cb_variable.activated[int].emit(1)
        self.assertTrue(cb_connector.isEnabled())
        self.assertEqual(widget.connect_value, 0)
        self.assertEqual(widget.connector_value, 0)

        widget.variable = a
        cb_variable.activated[int].emit(0)
        self.assertFalse(cb_connector.isEnabled())
        self.assertEqual(widget.connect_value, 0)
        self.assertEqual(widget.connector_value, 2)

    def test_callbacks_called_on_value(self):
        widget = self.widget
        send = widget.Outputs.network.send = Mock()
        update = widget.update_output = Mock(side_effect=widget.update_output)

        self._set_graph(Table.from_domain(Domain([self.c])))
        update.assert_called()
        update.reset_mock()
        send.assert_called()
        send.reset_mock()

        widget.connect_value = 1
        widget.controls.connect_value.activated[int].emit(1)
        update.assert_called()
        update.reset_mock()
        send.assert_called()
        send.reset_mock()

        widget.connector_value = 1
        widget.controls.connector_value.activated[int].emit(1)
        update.assert_called()
        update.reset_mock()
        send.assert_called()
        send.reset_mock()

    @patch("orangecontrib.network.network.twomode.to_single_mode")
    def test_masks_in_update(self, to_single_mode):
        widget = self.widget

        def check_call(expected_mode_mask, expected_conn_mask):
            to_single_mode.assert_called()
            net, mode_mask, conn_mask, weighting = to_single_mode.call_args[0]
            self.assertIs(net, widget.network)
            np.testing.assert_almost_equal(mode_mask, expected_mode_mask)
            np.testing.assert_almost_equal(conn_mask, expected_conn_mask)
            self.assertEqual(weighting, widget.weighting)

        self._set_graph(self.table)

        widget.variable = self.a
        widget.connect_value = 0
        widget.connector_value = 0
        widget.update_output()
        check_call(
            np.array([True, False, False, True, True]),
            np.array([False, True, True, False, False]))
        widget.connector_value = 2
        widget.update_output()
        check_call(
            np.array([True, False, False, True, True]),
            np.array([False, True, True, False, False]))
        widget.connect_value = 1
        widget.connector_value = 0
        widget.update_output()
        check_call(
            np.array([False, True, True, False, False]),
            np.array([True, False, False, True, True]))
        widget.connector_value = 2
        widget.update_output()
        check_call(
            np.array([False, True, True, False, False]),
            np.array([True, False, False, True, True]))

        widget.variable = self.c
        widget.connect_value = 0
        widget.connector_value = 0
        widget.update_output()
        check_call(
            np.array([False] * 5),
            np.array([True] * 5))

        widget.connect_value = 1
        widget.connector_value = 0
        widget.update_output()
        check_call(
            np.array([False, False, True, True, True]),
            np.array([True, True, False, False, False]))

        widget.connect_value = 1
        widget.connector_value = 1
        widget.update_output()
        check_call(
            np.array([False, False, True, True, True]),
            np.array([False, False, False, False, False]))

        widget.connect_value = 1
        widget.connector_value = 3
        widget.update_output()
        check_call(
            np.array([False, False, True, True, True]),
            np.array([True, False, False, False, False]))

        widget.connect_value = 1
        widget.connector_value = 4
        widget.update_output()
        check_call(
            np.array([False, False, True, True, True]),
            np.array([False, True, False, False, False]))

        widget.connect_value = 2
        widget.connector_value = 0
        widget.update_output()
        check_call(
            np.array([True, False, False, False, False]),
            np.array([False, True, True, True, True]))

        widget.connect_value = 2
        widget.connector_value = 1
        widget.update_output()
        check_call(
            np.array([True, False, False, False, False]),
            np.array([False, False, False, False, False]))

        widget.connect_value = 2
        widget.connector_value = 2
        widget.update_output()
        check_call(
            np.array([True, False, False, False, False]),
            np.array([False, False, True, True, True]))

        widget.connect_value = 2
        widget.connector_value = 4
        widget.update_output()
        check_call(
            np.array([True, False, False, False, False]),
            np.array([False, True, False, False, False]))

        widget.weighting = 3
        widget.update_output()
        check_call(
            np.array([True, False, False, False, False]),
            np.array([False, True, False, False, False]))

    def test_send_report(self):
        self._set_graph(self.table)
        self.widget.send_report()

    def test_missing_data(self):
        widget = self.widget
        network = self._read_network("davis.net")
        num_total = len(network.nodes)
        with network.nodes.unlocked(network.nodes.X):
            network.nodes.X[0, 1] = np.nan  # hide a node's role (= person in this case)
        self.send_signal(widget.Inputs.network, network)

        # Feature: role, Connect: person
        widget.variable = network.nodes.domain.attributes[1]
        widget.connect_value = 1
        widget.connector_value = 1
        widget.update_output()
        self.assertTrue(widget.Warning.ignoring_missing.is_shown())
        person_network = self.get_output(widget.Outputs.network)
        num_persons = len(person_network.nodes)

        # Feature: role, Connect: event
        widget.connect_value = 0
        widget.connector_value = 2
        widget.update_output()
        event_network = self.get_output(widget.Outputs.network)
        num_events = len(event_network.nodes)

        # Make sure the node with missing data is ignored for both modes
        self.assertEqual(num_persons + num_events, num_total - 1)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import os
from distutils.core import Extension

import numpy
from Cython.Build import cythonize
from setuptools import find_namespace_packages, setup
from setuptools.command.install import install

NAME = 'Orange3-Network'
DOCUMENTATION_NAME = 'Orange Network'

VERSION = '1.10.1'

DESCRIPTION = 'Networks add-on for Orange 3 data mining software package.'
LONG_DESCRIPTION  = open(os.path.join(os.path.dirname(__file__),
                                      'README.pypi')).read()
AUTHOR = 'Laboratory of Bioinformatics, FRI UL'
AUTHOR_EMAIL = 'info@biolab.si'
URL = 'https://github.com/biolab/orange3-network'
LICENSE = 'GPLv3'

KEYWORDS = [
    'network',
    'network analysis',
    'network layout',
    'network visualization',
    'data mining',
    'machine learning',
    'artificial intelligence',
    'orange',
    'orange3 add-on',
]

CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Environment :: X11 Applications :: Qt',
    'Environment :: Console',
    'Environment :: Plugins',
    'Programming Language :: Python',
    'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
    'Operating System :: OS Independent',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Topic :: Scientific/Engineering :: Visualization',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Intended Audience :: Education',
    'Intended Audience :: Science/Research',
    'Intended Audience :: Developers',
]

PACKAGES = find_namespace_packages()

PACKAGE_DATA = {
    'orangecontrib.network': ['networks/*'],
    'orangecontrib.network.widgets': ['icons/*'],
    'orangecontrib.network.widgets.tests': ['networks/*'],
    'orangecontrib.network.tests': ['*']
}

DATA_FILES = [
    # Data files that will be installed outside site-packages folder
]

SETUP_REQUIRES = (
    'setuptools',
    'trubar>=0.3.3',
)

INSTALL_REQUIRES = (
    'anyqt',
    'gensim',
    'Orange3>=3.38.1',
    'orange-widget-base',
    'pyqtgraph',
    'scipy',
    'scikit-learn',
),

EXTRAS_REQUIRE = {
    # Dependencies which are problematic to install automatically
    'GUI': (
        'AnyQt',
    ),
    'reST': (
        'numpydoc',
    ),
    'test': (
        'coverage',
    ),
    'doc': (
        'sphinx', 'recommonmark', 'sphinx_rtd_theme'
    ),
}

DEPENDENCY_LINKS = (
)

ENTRY_POINTS = {
    'orange.addons': (
        'network = orangecontrib.network',
    ),
    'orange.widgets': (
        'Network = orangecontrib.network.widgets',
    ),
    'orange.data.io.search_paths': (
        'network = orangecontrib.network:networks',
    ),
    # Register widget help
    "orange.canvas.help": (
        'html-index = orangecontrib.network.widgets:WIDGET_HELP_PATH',
    )
}

NAMESPACES = ['orangecontrib']
EXTENSIONS = [
    Extension(
        "*",
        ["orangecontrib/network/network/layout/*.pyx"],
        include_dirs=[numpy.get_include()],
        libraries=["m"] if os.name == 'posix' else [],
    )
]


def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration(None)
    config.set_options(ignore_setup_xxx_py=True,
                       assume_default_configuration=True,
                       delegate_options_to_subpackages=True,
                       quiet=True)
    config.add_subpackage('orangecontrib.network')
    return config


def include_documentation(local_dir, install_dir):
    global DATA_FILES

    doc_files = []
    for dirpath, _, files in os.walk(local_dir):
        doc_files.append(
            (
                dirpath.replace(local_dir, install_dir),
                [os.path.join(dirpath, f) for f in files]
            )
        )
    DATA_FILES.extend(doc_files)


class InstallMultilingualCommand(install):
    def run(self):
        install.run(self)
        self.compile_to_multilingual()

    def compile_to_multilingual(self):
        from trubar import translate

        package_dir = os.path.dirname(os.path.abspath(__file__))
        translate(
            "msgs.jaml",
            source_dir=os.path.join(self.install_lib, "orangecontrib", "network"),
            config_file=os.path.join(package_dir, "i18n", "trubar-config.yaml"))


if __name__ == '__main__':
    cmdclass = {}

    include_documentation('doc/_build/html', 'help/orange3-network')

    setup(
        configuration=configuration,
        name=NAME,
        version=VERSION,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type='text/markdown',
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        url=URL,
        license=LICENSE,
        keywords=KEYWORDS,
        classifiers=CLASSIFIERS,
        packages=PACKAGES,
        ext_modules=cythonize(EXTENSIONS),
        package_data=PACKAGE_DATA,
        data_files=DATA_FILES,
        python_requires=">=3.8",
        setup_requires=SETUP_REQUIRES,
        cmdclass={
            'install': InstallMultilingualCommand,
        },
        install_requires=INSTALL_REQUIRES,
        extras_require=EXTRAS_REQUIRE,
        dependency_links=DEPENDENCY_LINKS,
        entry_points=ENTRY_POINTS,
        namespace_packages=NAMESPACES,
        include_package_data=True,
        zip_safe=False,
    )



"""
.. index:: Community Detection in Graphs

.. index::
   single: Network; Community Detection in Graphs

*****************************
Community Detection in Graphs
*****************************

"""

import random
import itertools
import numpy as np

from Orange.data import Domain, Table, DiscreteVariable


def add_results_to_items(G, labels, var_name):
    items = G.nodes

    new_attr = DiscreteVariable(
        var_name, values=["C%d" % (x + 1) for x in set(labels.values())])
    new_data = np.array(list(labels.values())).reshape(-1, 1)

    if items is not None:
        # a unique `var_name` is assumed to be obtained in OWNxClustering prior
        # to calling this; only check metas to enable overriding results on
        # multiple reruns
        effective_metas = [idx for idx, meta in enumerate(items.domain.metas)
                           if meta.name != var_name]
        metas = np.take(items.domain.metas, effective_metas).tolist()
        metas.append(new_attr)

        meta_data = items.metas[:, effective_metas]
        meta_data = np.hstack((meta_data, new_data))

        domain = Domain(items.domain.attributes, items.domain.class_vars, metas)
        data = Table(domain, items.X, items.Y, meta_data, items.W)
        G.nodes = data
    else:
        empty = np.array([[] for _ in range(new_data.shape[0])])
        domain = Domain([], [], metas=[new_attr])
        data = Table(domain, empty, empty, new_data)
        G.node = data


class CommunityDetection(object):
    def __init__(self, algorithm, **kwargs):
        self.algorithm = algorithm
        self.kwargs = kwargs

    def __call__(self, G):
        return self.algorithm(G, **self.kwargs)


def label_propagation_hop_attenuation(G, iterations=1000,
                                      delta=0.1, node_degree_preference=0, seed=None):
    """Label propagation for community detection, Leung et al., 2009

    :param G: A Orange graph.
    :type G: Orange.network.Graph

    :param iterations: The max. number of iterations if no convergence.
    :type iterations: int

    :param delta: The hop attenuation factor.
    :type delta: float

    :param node_degree_preference: The power on node degree factor.
    :type node_degree_preference: float

    :param seed: Seed for replicable clustering.
    :type seed: int
    """

    if G.edges[0].directed:
        raise ValueError("Undirected graph expected")

    if seed is not None:
        random.seed(seed)

    vertices = list(range(G.number_of_nodes()))
    degrees = dict(enumerate(G.degrees()))
    labels = dict(zip(vertices, range(G.number_of_nodes())))
    scores = dict(zip(vertices, [1] * G.number_of_nodes()))
    m = node_degree_preference

    for i in range(iterations):
        random.shuffle(vertices)
        stop = 1
        for v in vertices:
            neighbors = list(G.neighbours(v))
            if len(neighbors) == 0:
                continue

            lbls = sorted(((G.edges[0].edges[v, u], labels[u], u)
                           for u in neighbors), key=lambda x: x[1])
            lbls = [(sum(scores[u] * degrees[u] ** m * weight for weight,
                         _u_label, u in group), label)
                    for label, group in itertools.groupby(lbls, lambda x: x[1])]
            max_score = max(lbls)[0]
            max_lbls = [label for score, label in lbls if score >= max_score]

            # only change label if it is not already one of the
            # preferred (maximal) labels
            if labels[v] not in max_lbls:
                labels[v] = random.choice(max_lbls)
                scores[v] = max(0, max(scores[u] for u in neighbors \
                                       if labels[u] == labels[v]) - delta)
                stop = 0

        # if stopping condition is satisfied (no label switched color)
        if stop:
            break

    labels = remap_labels(labels)

    return labels


def label_propagation(G, iterations=1000, seed=None):
    """Label propagation for community detection, Raghavan et al., 2007

    :param G: A Orange graph.
    :type G: Orange.network.Graph

    :param iterations: The maximum number of iterations if there is no convergence.
    :type iterations: int

    :param seed: Seed for replicable clustering.
    :type seed: int
    """
    if seed is not None:
        random.seed(seed)

    vertices = list(range(G.number_of_nodes()))
    labels = dict(zip(vertices, range(G.number_of_nodes())))

    def next_label(neighbors):
        """Updating rule as described by Raghavan et al., 2007

        Return a list of possible node labels with equal probability.
        """
        lbls = sorted(labels[u] for u in neighbors)
        lbls = [(len(list(c)), l) for l, c in itertools.groupby(lbls)]
        m = max(lbls)[0]
        return [l for c, l in lbls if c >= m]

    for i in range(iterations):
        random.shuffle(vertices)
        stop = 1
        for v in vertices:
            nbh = list(G.neighbours(v))
            if len(nbh) == 0:
                continue

            max_lbls = next_label(nbh)

            if labels[v] not in max_lbls:
                stop = 0

            labels[v] = random.choice(max_lbls)

        # if stopping condition might be satisfied, check it
        # stop when no label would switch anymore
        if stop:
            for v in vertices:
                nbh = list(G.neighbours(v))
                if len(nbh) == 0:
                    continue
                max_lbls = next_label(nbh)
                if labels[v] not in max_lbls:
                    stop = 0
                    break

            if stop:
                break

    labels = remap_labels(labels)

    return labels


def remap_labels(labels):
    unique_labels = sorted(set(labels.values()))
    dict_labels = {label:i for i, label in enumerate(unique_labels)}
    return {key: dict_labels[val] for key, val in labels.items()}



import unittest

import numpy as np
import scipy.sparse as sp
from Orange.data import Table, ContinuousVariable, Domain

from orangecontrib.network import Network
from orangecontrib.network.network.base import DirectedEdges, UndirectedEdges
from orangecontrib.network.network.embeddings import Node2Vec


class TestEmbeddings(unittest.TestCase):
    def setUp(self):
        row, col, w = zip(*((1, 2, 1.0), (1, 3, 3.0), (2, 3, 1.0), (2, 6, 0.5), (3, 4, 1.0), (4, 5, 1.0), (4, 7, -1.0),
                            (5, 6, 0.0), (6, 5, 0.1), (6, 2, 0.1)))
        dir_edges = DirectedEdges(sp.csr_matrix((w, (row, col)), shape=(8, 8)))
        self.toy_directed = Network(np.arange(8), dir_edges)

        row, col, w = zip(*((1, 2, 1.0), (1, 3, 3.0), (2, 3, 1.0), (2, 6, 0.5), (3, 4, 1.0), (4, 5, 1.0), (4, 7, -1.0),
                            (5, 6, 0.1)))
        undir_edges = UndirectedEdges(sp.csr_matrix((w, (row, col)), shape=(8, 8)))
        self.toy_undirected = Network(np.arange(8), undir_edges)

    def test_node_probas(self):
        """ Test that node probabilities get calculated correctly """
        n2v = Node2Vec()
        # nowhere to go from isolated node
        self.assertEqual(len(n2v.node_probas(self.toy_directed, 0)), 0)
        # should not have division by zero when weights of edges to neighbors sum to 0
        self.assertDictEqual(n2v.node_probas(self.toy_directed, 5), {6: 1.0})
        probas = n2v.node_probas(self.toy_directed, 4)
        self.assertAlmostEqual(probas[5], 0.881, places=3)
        self.assertAlmostEqual(probas[7], 0.119, places=3)

        self.assertDictEqual(n2v.node_probas(self.toy_directed, 1), {2: 0.25, 3: 0.75})
        self.assertDictEqual(n2v.node_probas(self.toy_undirected, 3), {1: 0.6, 2: 0.2, 4: 0.2})

    def test_edge_probas(self):
        """ Test that edge probabilities get calculated appropriately based on shortest distance between previous node
            and next node (equations in 'Search bias' section of node2vec paper) """
        n2v = Node2Vec(p=0.8, q=0.5)
        edge_probas = n2v.edge_probas(self.toy_directed, 3, 4)
        self.assertAlmostEqual(edge_probas[(4, 5)], 0.982, places=3)  # d_tx = 2
        self.assertAlmostEqual(edge_probas[(4, 7)], 0.018, places=3)  # d_tx = 2
        edge_probas = n2v.edge_probas(self.toy_directed, 1, 2)
        self.assertAlmostEqual(edge_probas[(2, 3)], 0.5)  # d_tx = 1
        edge_probas = n2v.edge_probas(self.toy_directed, 5, 6)
        self.assertAlmostEqual(edge_probas[(6, 5)], 0.385, places=3)  # d_tx = 0

        edge_probas = n2v.edge_probas(self.toy_undirected, 1, 2)
        self.assertAlmostEqual(edge_probas[(2, 1)], 0.385, places=3)  # d_tx = 0
        self.assertAlmostEqual(edge_probas[(2, 3)], 0.308, places=3)  # d_tx = 1
        self.assertAlmostEqual(edge_probas[(2, 6)], 0.308, places=3)  # d_tx = 2

    def test_call(self):
        n2v = Node2Vec(num_walks=10, walk_len=80, emb_size=300)
        embeddings = n2v(self.toy_directed)
        self.assertEqual(embeddings.X.shape, (self.toy_directed.number_of_nodes(), 300))

        # check that domain is extended and that the  existing attributes do not change places
        empty = np.array([[] for _ in range(8)])
        data = Table(Domain([ContinuousVariable("var1")]), np.array([[i] for i in range(8)]), empty, empty)
        toy_net_with_data = Network(data, self.toy_directed.edges[0])
        extended_data = n2v(toy_net_with_data)

        self.assertEqual(extended_data.X.shape, (toy_net_with_data.number_of_nodes(), 1 + 300))
        np.testing.assert_array_almost_equal(extended_data.X[:, 0], np.arange(8))


if __name__ == '__main__':
    unittest.main()



import numpy as np

from Orange.data import Table, Domain, DiscreteVariable, ContinuousVariable

from orangecontrib.network.network import generate
from orangecontrib.network.widgets.OWNxClustering import OWNxClustering
from orangecontrib.network.widgets.tests.utils import NetworkTest


class TestOWNxClustering(NetworkTest):
    def setUp(self):
        self.widget = self.create_widget(
            OWNxClustering, stored_settings={'autoApply': False},
        )  # type: OWNxClustering

    def test_does_not_crash_when_cluster_variable_already_exists(self):
        # Prepare some dummy data
        x, clusters = np.ones((5, 1)), np.ones((5, 1))

        data_var = ContinuousVariable('Data')
        cluster_var = DiscreteVariable('Cluster', ('C1', 'C2'))
        data = Table.from_numpy(
            Domain([data_var], metas=[cluster_var]),
            X=x, metas=clusters,
        )
        graph = generate.complete(5)
        graph.nodes = data

        # Should not crash
        self.send_signal(self.widget.Inputs.network, graph)
        self.widget.unconditional_commit()

        # There should be an additional cluster column
        output = self.get_output(self.widget.Outputs.network).nodes
        self.assertEqual(len(data.domain.metas) + 1, len(output.domain.metas))

    def test_reproducible_clustering(self):
        network = self._read_network("leu_by_genesets.net")
        self.widget.controls.use_random_state.setChecked(True)

        self.send_signal(self.widget.Inputs.network, network)
        self.widget.unconditional_commit()
        res1 = self.get_output(self.widget.Outputs.network).nodes.metas

        self.send_signal(self.widget.Inputs.network, network)
        self.widget.unconditional_commit()
        res2 = self.get_output(self.widget.Outputs.network).nodes.metas

        # Seeded rerun should give same clustering result
        self.assertTrue(np.all(res1 == res2))

    def test_multiple_reruns_override_variable(self):
        network = self._read_network("leu_by_genesets.net")
        num_initial_metas = len(network.nodes.domain.metas)

        for _ in range(10):
            self.send_signal(self.widget.Inputs.network, network)
            self.widget.unconditional_commit()

        output = self.get_output(self.widget.Outputs.network).nodes
        # Multiple reruns should override the results instead of adding new feature every time
        self.assertEqual(len(output.domain.metas), num_initial_metas + 1)



import numpy as np
import scipy.sparse as sp

from Orange.data import DiscreteVariable, Table, Domain
from Orange.widgets import gui
from Orange.widgets.settings import ContextSetting, DomainContextHandler, \
    Setting
from Orange.widgets.utils.itemmodels import DomainModel
from Orange.widgets.widget import Input, Output, OWWidget, Msg
from orangecontrib.network import Network


class OWNxGroups(OWWidget):
    name = "Network Of Groups"
    description = "Group instances by feature and connect related groups."
    icon = "icons/NetworkGroups.svg"
    priority = 6435

    class Inputs:
        network = Input("Network", Network, default=True)
        data = Input("Data", Table)

    class Outputs:
        network = Output("Network", Network, default=True)
        data = Output("Data", Table)

    class Warning(OWWidget.Warning):
        no_graph_found = Msg("Data is given, network is missing.")
        no_discrete_features = Msg("Data has no discrete features.")

    class Error(OWWidget.Error):
        data_size_mismatch = Msg("Length of the data does not "
                                 "match the number of nodes.")

    resizing_enabled = False
    want_main_area = False

    NoWeights, WeightByDegrees, WeightByWeights = range(3)
    weight_labels =\
        ["No weights", "Number of connections", "Sum of connection weights"]

    settingsHandler = DomainContextHandler()
    feature = ContextSetting(None)
    weighting = Setting(2)
    normalize = Setting(True)

    def __init__(self):
        super().__init__()
        self.network = None
        self.data = None
        self.effective_data = None
        self.out_nodes = self.out_edges = None

        info_box = gui.widgetBox(self.controlArea, "Info")
        self.input_label = gui.widgetLabel(info_box, "")
        self.output_label = gui.widgetLabel(info_box, "")
        self._set_input_label_text()
        self._set_output_label_text(None)

        gui.comboBox(
            self.controlArea, self, "feature", box="Group by",
            callback=self.__feature_combo_changed,
            model=DomainModel(valid_types=DiscreteVariable)
        )
        radios = gui.radioButtons(
            self.controlArea, self, "weighting", box="Output weights",
            btnLabels=self.weight_labels, callback=self.__feature_combo_changed
        )
        gui.checkBox(
            gui.indentedBox(radios),
            self, "normalize", "Normalize by geometric mean",
            callback=self.__normalization_changed
        )
        self.controls.normalize.setEnabled(
            self.weighting == self.WeightByWeights)

    def _set_input_label_text(self):
        if self.network is None:
            self.input_label.setText("Input: no data")
        else:
            self.input_label.setText(
                f"Input: "
                f"{self.network.number_of_nodes()} nodes, "
                f"{self.network.number_of_edges()} edges")

    def _set_output_label_text(self, output_network):
        if output_network is None:
            self.output_label.setText("Output: no data")
            self.out_nodes = self.out_edges = None
        else:
            self.out_nodes = output_network.number_of_nodes()
            self.out_edges = output_network.number_of_edges()
            self.output_label.setText(
                f"Output: {self.out_nodes} nodes, {self.out_edges} edges"
            )

    def __feature_combo_changed(self):
        self.controls.normalize.setEnabled(
            self.weighting == self.WeightByWeights)
        self.commit()

    def __normalization_changed(self):
        self.commit()

    @Inputs.network
    def set_network(self, network):
        self.network = network
        self._set_input_label_text()

    @Inputs.data
    def set_data(self, data):
        self.data = data

    def handleNewSignals(self):
        self.closeContext()
        self.clear_messages()
        self.set_effective_data()
        self.set_feature_model()
        if self.controls.feature.model():
            self.openContext(self.effective_data)
        self.commit()

    def set_effective_data(self):
        self.effective_data = None
        if self.network is None and self.data is not None:
            self.Warning.no_graph_found()
        elif self.network is not None and self.data is not None:
            if len(self.data) != self.network.number_of_nodes():
                self.Error.data_size_mismatch()
            else:
                self.effective_data = self.data
        elif self.data is None and self.network is not None:
            self.effective_data = self.network.nodes

        if self.effective_data is not None and not \
                self.effective_data.domain.has_discrete_attributes(True, True):
            self.Warning.no_discrete_features()

    def set_feature_model(self):
        data = self.effective_data
        feature_model = self.controls.feature.model()
        feature_model.set_domain(data.domain if data else None)
        self.feature = feature_model[0] if feature_model else None

    def commit(self):
        if self.feature is None:
            output_network = None
        else:
            output_network = self._map_network()
        self.Outputs.network.send(output_network)
        self.Outputs.data.send(output_network and output_network.nodes)
        self._set_output_label_text(output_network)

    def _map_network(self):
        edges = self.network.edges[0].edges.tocoo()
        row, col = edges.row, edges.col
        if self.weighting == self.WeightByWeights:
            weights = edges.data
        else:
            weights = None
        if self.normalize:
            self._normalize_weights(row, col, weights)
        row, col = self._map_into_feature_values(row, col)
        return Network(
            self._construct_items(),
            self._construct_edges(row, col, weights))

    def _normalize_weights(self, row, col, weights):
        if weights is None:
            weights = np.ones((len(row)), dtype=float)
            degs = self.network.degrees()
        else:
            degs = self.network.degrees(weighted=True)
        weights /= np.sqrt(degs.T[row] * degs.T[col])

    def _map_into_feature_values(self, row, col):
        selected_column = self.effective_data.get_column(self.feature)
        return (selected_column[row].astype(np.float64),
                selected_column[col].astype(np.float64))

    def _construct_edges(self, col, row, weights):
        # remove edges that connect to "unknown" group
        mask = ~np.any(np.isnan(np.vstack((row, col))), axis=0)
        # remove edges within a node
        mask = np.logical_and((row != col), mask)
        row, col = row[mask], col[mask]
        if weights is not None:
            weights = weights[mask]

        # find unique edges
        mask = row > col
        row[mask], col[mask] = col[mask], row[mask]

        array = np.vstack((row.astype(int), col.astype(int)))
        (row, col), inverse = np.unique(array, axis=1, return_inverse=True)

        if self.weighting == self.NoWeights:
            data = np.ones(len(row))
        elif self.weighting == self.WeightByDegrees:
            data = np.fromiter(
                (np.sum(inverse == i).astype(float) for i in range(len(row))),
                dtype=float, count=len(row))
        else:
            data = np.fromiter(
                (np.sum(weights[inverse == i]) for i in range(len(row))),
                dtype=float, count=len(row))
        dim = len(self.feature.values)
        return sp.csr_matrix((data, (row, col)), shape=(dim, dim))

    def _construct_items(self):
        domain = Domain([self.feature])
        return Table(domain, np.arange(len(self.feature.values))[:, None])

    def send_report(self):
        if not self.effective_data:
            return

        self.report_items("Input network", [
            ("Number of vertices", self.network.number_of_nodes()),
            ("Number of edges", self.network.number_of_edges())])
        self.report_data("Input data", self.effective_data)
        self.report_items("Settings", [
            ("Group by", self.feature.name),
            ("Weights", self.weight_labels[self.weighting].lower() +
             (", normalized by geometric mean" if self.normalize else ""))
        ])
        if self.out_nodes is not None:
            self.report_items("Output network", [
                ("Number of vertices", self.out_nodes),
                ("Number of edges", self.out_edges)])


def main():  # pragma: no cover
    from orangecontrib.network.network.readwrite import read_pajek
    from os.path import join, dirname
    from orangewidget.utils.widgetpreview import WidgetPreview

    path = join(dirname(__file__), "..", "networks")
    network = read_pajek(join(path, 'airtraffic.net'))
    data = Table(join(path, 'airtraffic_items.tab'))

    WidgetPreview(OWNxGroups).run(set_network=network, set_data=data)


if __name__ == "__main__":  # pragma: no cover
    main()
